# pyright: reportPrivateUsage=false
import sys
from argparse import ArgumentError
from collections.abc import MutableSequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pytest

from dictIO.cli.dict_parser import _argparser, _validate_scope, main
from dictIO.dict_parser import DictParser

# *****Test commandline interface (CLI)************************************************************


@dataclass()
class CliArgs:
    # Expected default values for the CLI arguments when dictParser gets called via the commandline
    quiet: bool = False
    verbose: bool = False
    log: str | None = None
    log_level: str = "WARNING"

    dict: str | None = "test_dictParser_dict"
    ignore_includes: bool = False
    mode: str = "w"
    order: bool = False
    ignore_comments: bool = False
    scope: str | None = None
    output: str = "cpp"


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], ArgumentError),
        (["test_dictParser_dict"], CliArgs()),
        (["test_dictParser_dict", "-q"], CliArgs(quiet=True)),
        (["test_dictParser_dict", "--quiet"], CliArgs(quiet=True)),
        (["test_dictParser_dict", "-v"], CliArgs(verbose=True)),
        (["test_dictParser_dict", "--verbose"], CliArgs(verbose=True)),
        (["test_dictParser_dict", "-qv"], ArgumentError),
        (["test_dictParser_dict", "--log", "logFile"], CliArgs(log="logFile")),
        (["test_dictParser_dict", "--log"], ArgumentError),
        (["test_dictParser_dict", "--log-level", "INFO"], CliArgs(log_level="INFO")),
        (["test_dictParser_dict", "--log-level"], ArgumentError),
        (["test_dictParser_dict", "-I"], CliArgs(ignore_includes=True)),
        (["test_dictParser_dict", "--ignore-includes"], CliArgs(ignore_includes=True)),
        (["test_dictParser_dict", "--mode", "a"], CliArgs(mode="a")),
        (["test_dictParser_dict", "--mode"], ArgumentError),
        (["test_dictParser_dict", "--order"], CliArgs(order=True)),
        (["test_dictParser_dict", "-o"], ArgumentError),
        (["test_dictParser_dict", "-C"], CliArgs(ignore_comments=True)),
        (["test_dictParser_dict", "--ignore-comments"], CliArgs(ignore_comments=True)),
        (["test_dictParser_dict", "--scope", ""], CliArgs(scope="")),
        (["test_dictParser_dict", "--scope", "key"], CliArgs(scope="key")),
        (
            ["test_dictParser_dict", "--scope", "[key1,key2]"],
            CliArgs(scope="[key1,key2]"),
        ),
        (
            ["test_dictParser_dict", "--scope", "[key1, key2]"],
            CliArgs(scope="[key1, key2]"),
        ),
        (
            ["test_dictParser_dict", "--scope", "['key1', 'key2']"],
            CliArgs(scope="['key1', 'key2']"),
        ),
        (["test_dictParser_dict", "--scope", "key1", "key2"], ArgumentError),
        (["test_dictParser_dict", "--scope"], ArgumentError),
        (["test_dictParser_dict", "--output", "cpp"], CliArgs(output="cpp")),
        (["test_dictParser_dict", "--output", "foam"], CliArgs(output="foam")),
        (["test_dictParser_dict", "--output", "xml"], CliArgs(output="xml")),
        (["test_dictParser_dict", "--output", "json"], CliArgs(output="json")),
        (["test_dictParser_dict", "--output", ""], ArgumentError),
        (["test_dictParser_dict", "--output"], ArgumentError),
    ],
)
def test_cli(
    inputs: list[str],
    expected: CliArgs | type,
    monkeypatch: pytest.MonkeyPatch,
):
    # sourcery skip: no-conditionals-in-tests
    # sourcery skip: no-loop-in-tests
    # Prepare
    monkeypatch.setattr(sys, "argv", ["dictParser", *inputs])
    parser = _argparser()
    # Execute
    if isinstance(expected, CliArgs):
        args_expected: CliArgs = expected
        args = parser.parse_args()
        # Assert args
        for key in args_expected.__dataclass_fields__:
            assert args.__getattribute__(key) == args_expected.__getattribute__(key)
    elif issubclass(expected, Exception):
        exception: type = expected
        # Assert that expected exception is raised
        with pytest.raises((exception, SystemExit)):
            args = parser.parse_args()
    else:
        raise AssertionError


# *****Ensure the CLI correctly invokes the API****************************************************


@dataclass()
class ApiArgs:
    # Values that main() is expected to pass to DictParser.parse() by default when invoking the API
    source_file: Path = Path("test_dictParser_dict")
    includes: bool = True
    mode: str = "w"
    order: bool = False
    comments: bool = True
    scope: MutableSequence[Any] | None = None
    output: str | None = "cpp"


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], ArgumentError),
        (["test_dictParser_dict"], ApiArgs()),
        (["test_dictParser_dict", "-I"], ApiArgs(includes=False)),
        (["test_dictParser_dict", "--ignore-includes"], ApiArgs(includes=False)),
        (["test_dictParser_dict", "--mode", "a"], ApiArgs(mode="a")),
        (["test_dictParser_dict", "--mode"], ArgumentError),
        (["test_dictParser_dict", "--order"], ApiArgs(order=True)),
        (["test_dictParser_dict", "-o"], ArgumentError),
        (["test_dictParser_dict", "-C"], ApiArgs(comments=False)),
        (["test_dictParser_dict", "--ignore-comments"], ApiArgs(comments=False)),
        (["test_dictParser_dict", "--scope", ""], ApiArgs(scope=[""])),
        (["test_dictParser_dict", "--scope", "key"], ApiArgs(scope=["key"])),
        (
            ["test_dictParser_dict", "--scope", "[key1,key2]"],
            ApiArgs(scope=["key1", "key2"]),
        ),
        (
            ["test_dictParser_dict", "--scope", "[key1, key2]"],
            ApiArgs(scope=["key1", "key2"]),
        ),
        (
            ["test_dictParser_dict", "--scope", "['key1', 'key2']"],
            ApiArgs(scope=["key1", "key2"]),
        ),
        (["test_dictParser_dict", "--scope", "key1", "key2"], ArgumentError),
        (["test_dictParser_dict", "--scope"], ArgumentError),
        (["test_dictParser_dict", "--output", "cpp"], ApiArgs(output="cpp")),
        (["test_dictParser_dict", "--output", "foam"], ApiArgs(output="foam")),
        (["test_dictParser_dict", "--output", "xml"], ApiArgs(output="xml")),
        (["test_dictParser_dict", "--output", "json"], ApiArgs(output="json")),
        (["test_dictParser_dict", "--output", ""], ArgumentError),
        (["test_dictParser_dict", "--output"], ArgumentError),
    ],
)
def test_invoke_api(
    inputs: list[str],
    expected: ApiArgs | type,
    monkeypatch: pytest.MonkeyPatch,
):
    # sourcery skip: no-conditionals-in-tests
    # sourcery skip: no-loop-in-tests
    # Prepare
    monkeypatch.setattr(sys, "argv", ["dictParser", *inputs])
    args: ApiArgs = ApiArgs()

    def fake_parse(
        source_file: Path,
        *,
        includes: bool = True,
        mode: str = "w",
        order: bool = False,
        comments: bool = True,
        scope: MutableSequence[Any] | None = None,
        output: str | None = None,
    ):
        args.source_file = source_file
        args.includes = includes
        args.mode = mode
        args.order = order
        args.comments = comments
        args.scope = scope
        args.output = output

    monkeypatch.setattr(target=DictParser, name="parse", value=fake_parse)
    # Execute
    if isinstance(expected, ApiArgs):
        args_expected: ApiArgs = expected
        main()
        # Assert args
        for key in args_expected.__dataclass_fields__:
            assert args.__getattribute__(key) == args_expected.__getattribute__(key)
    elif issubclass(expected, Exception):
        exception: type = expected
        # Assert that expected exception is raised
        with pytest.raises((exception, SystemExit)):
            main()
    else:
        raise AssertionError


# *****Test _validate_scope() helper function******************************************************


def test_validate_scope() -> None:
    # Vary scope to different types and each time test whether setOptions works as expected
    scope_in: str | list[str] | int | None
    # None
    scope_in = None
    scope_out: list[str] | list[str | int] | list[int] | None = _validate_scope(scope_in)
    assert scope_out is None

    # empty string -> should be returned as a one-element list
    scope_in = ""
    scope_out = _validate_scope(scope_in)
    assert isinstance(scope_out, list)
    assert len(scope_out) == 1
    assert list(scope_out)[0] == ""

    # (single) string -> should be returned as a one-element list
    scope_in = "strings"
    scope_out = _validate_scope(scope_in)
    assert isinstance(scope_out, list)
    assert len(scope_out) == 1
    assert list(scope_out)[0] == "strings"

    # list
    scope_in = ["input", "time"]
    scope_out = _validate_scope(scope_in)
    assert scope_out is not None
    assert isinstance(scope_out, list)
    assert len(scope_out) == 2
    assert list(scope_out)[0] == "input"
    assert list(scope_out)[1] == "time"

    # list in string format (= string, which LOOKS like a list)
    scope_in = "['input','time']"
    scope_out = _validate_scope(scope_in)
    assert scope_out is not None
    assert isinstance(scope_out, list)
    assert len(scope_out) == 2
    assert list(scope_out)[0] == "input"
    assert list(scope_out)[1] == "time"

    # list in string format (= string, which LOOKS like a list)
    scope_in = "[input,time]"
    scope_out = _validate_scope(scope_in)
    assert scope_out is not None
    assert isinstance(scope_out, list)
    assert len(scope_out) == 2
    assert list(scope_out)[0] == "input"
    assert list(scope_out)[1] == "time"

    # list in string format (= string, which LOOKS like a list)
    scope_in = " [ input , time ] "
    scope_out = _validate_scope(scope_in)
    assert scope_out is not None
    assert isinstance(scope_out, list)
    assert len(scope_out) == 2
    assert list(scope_out)[0] == "input"
    assert list(scope_out)[1] == "time"

    # int  -> should not be accepted and result in None
    scope_in = 1
    scope_out = _validate_scope(scope_in)
    assert scope_out is None



#!/usr/bin/env python
# coding utf-8
"""dictParser command line interface."""

import argparse
import logging
import re
from collections.abc import MutableSequence
from pathlib import Path
from typing import Any, cast

from dictIO import DictParser
from dictIO.types import K
from dictIO.utils.logging import configure_logging

logger = logging.getLogger(__name__)


def _argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="dictParser",
        usage="%(prog)s dict [options [args]]",
        epilog="_________________dictParser___________________",
        prefix_chars="-",
        add_help=True,
        description=(
            "Reads a dict file, merges sub-dicts referenced through #include directives,\n"
            "evaluates variables and expressions,\n"
            "and finally saves the parsed dict with prefix 'parsed', i.e. parsed.<DICTNAME>."
            "The format of the output file will by default be dictIO native file format, but can optionally be\n"
            "changed to foam, xml or json format."
        ),
    )

    _ = parser.add_argument(
        "dict",
        metavar="dict",
        type=str,
        help="name of dict file to be parsed.",
    )

    _ = parser.add_argument(
        "-I",
        "--ignore-includes",
        action="store_true",
        help=("ignore include directives (e.g. #include './SUBDICT').\n" "This suppresses merging of sub-dicts. "),
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "--mode",
        help=(
            "'a' -- append to output file if a dict with the same name already exists; \n"
            "'w' -- overwrite output file (default)"
        ),
        choices=["a", "w"],
        default="w",
        required=False,
        type=str,
    )

    _ = parser.add_argument(
        "--order",
        action="store_true",
        help="sort the parsed dict.",
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "-C",
        "--ignore-comments",
        action="store_true",
        help="supress writing comments into the output file. The header is excepted and will always be written.",
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "--scope",
        action="store",
        help=(
            "optionally specify a scope the dict will be reduced to after parsing.\n"
            "'scope' can be EMPTY, a 'STRING' or a list of strings \"['STRING1', 'STRING2']\""
        ),
        default=None,
        required=False,
    )

    _ = parser.add_argument(
        "-o",
        "--output",
        action="store",
        type=str,
        help="specify format of the output file.",
        choices=["cpp", "foam", "xml", "json"],
        default="cpp",
        required=False,
    )

    console_verbosity = parser.add_mutually_exclusive_group(required=False)

    _ = console_verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=("console output will be quiet."),
        default=False,
    )

    _ = console_verbosity.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=("console output will be verbose."),
        default=False,
    )

    _ = parser.add_argument(
        "--log",
        action="store",
        type=str,
        help="name of log file. If specified, this will activate logging to file.",
        default=None,
        required=False,
    )

    _ = parser.add_argument(
        "--log-level",
        action="store",
        type=str,
        help="log level applied to logging to file.",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        required=False,
    )

    return parser


def main() -> None:
    """Entry point for console script as configured in setup.cfg.

    Runs the command line interface and parses arguments and options entered on the console.
    """
    parser = _argparser()
    args = parser.parse_args()

    # Configure Logging
    # ..to console
    log_level_console: str = "WARNING"
    if any([args.quiet, args.verbose]):
        log_level_console = "ERROR" if args.quiet else log_level_console
        log_level_console = "DEBUG" if args.verbose else log_level_console
    # ..to file
    log_file: Path | None = Path(args.log) if args.log else None
    log_level_file: str = args.log_level
    configure_logging(log_level_console, log_file, log_level_file)

    source_file = Path(args.dict)
    includes: bool = not args.ignore_includes
    mode: str = args.mode
    order: bool = args.order
    comments: bool = not args.ignore_comments
    # Validate scope: It needs to be a list of strings
    scope: MutableSequence[Any] | None = _validate_scope(args.scope)
    output: str | None = args.output

    # Check whether source file exists
    if not source_file.exists():
        logger.error(f"dictParser.py: File {source_file} not found.")
        return

    logger.info(
        f"Start dictParser.py with following arguments:\n"
        f"\t source_file: \t\t{source_file}\n"
        f"\t includes: \t\t\t{includes}\n"
        f"\t order: \t\t\t\t{order}\n"
        f"\t comments: \t\t\t{comments}\n"
        f"\t scope: \t\t\t\t{scope}\n"
        f"\t output: \t\t\t{output}"
    )

    # Invoke API
    if DictParser.parse(
        source_file=source_file,
        includes=includes,
        mode=mode,
        order=order,
        comments=comments,
        scope=scope,
        output=output,
    ):
        logger.info("dictParser.py finished successfully.\n")
    else:
        logger.error("dictParser.py finished with errors.\n")


def _validate_scope(
    scope: MutableSequence[K] | K | None,
) -> list[K] | None:
    # sourcery skip: replace-interpolation-with-fstring
    validated_scope: list[K] | None = None
    if isinstance(scope, MutableSequence):  # List
        validated_scope = list(scope)  # no conversion needed
    elif isinstance(scope, str):  # string
        if re.match(r"^\s*\[", scope):  # ..maybe a string that LOOKS like a list?
            try:  # Then try to convert that string to a list
                from dictIO import Parser

                parser = Parser()
                _scope: str = scope.strip(" []")
                validated_scope = cast(list[K], [key.strip() for key in _scope.split(",")])
                parser.parse_values(validated_scope)
            except Exception:
                logger.exception(f"setOptions: misspelled scope: {scope}")
        else:  # string is just a single value.
            # Store it not as string but as a (one-element) list
            validated_scope = [cast(K, scope)]
    else:  # 'scope' is neither a list nor a string -> set validated_scope to None
        validated_scope = None
    return validated_scope


if __name__ == "__main__":
    main()



"""dictIO package provides classes for reading, writing, and parsing dictionaries."""

from dictIO.dict import (
    SDict as SDict,
)

from dictIO.utils.dict import (
    order_keys as order_keys,
    find_global_key as find_global_key,
    set_global_key as set_global_key,
)

from dictIO.cpp_dict import (
    CppDict as CppDict,
)

from dictIO.formatter import (
    Formatter as Formatter,
    NativeFormatter as NativeFormatter,
    FoamFormatter as FoamFormatter,
    JsonFormatter as JsonFormatter,
    XmlFormatter as XmlFormatter,
)
from dictIO.parser import (
    Parser as Parser,
    NativeParser as NativeParser,
    FoamParser as FoamParser,
    JsonParser as JsonParser,
    XmlParser as XmlParser,
)

from dictIO.dict_reader import (
    DictReader as DictReader,
)
from dictIO.dict_writer import (
    DictWriter as DictWriter,
    create_target_file_name as create_target_file_name,
)
from dictIO.dict_parser import (
    DictParser as DictParser,
)



import copy
import re
from collections.abc import MutableMapping
from copy import deepcopy
from pathlib import Path
from typing import Any, cast

import pytest

from dictIO import (
    CppDict,
    DictParser,
    DictReader,
    DictWriter,
    NativeParser,
    SDict,
    find_global_key,
    order_keys,
    set_global_key,
)
from dictIO.types import K, V


@pytest.fixture
def test_dict() -> SDict[str, Any]:
    parser = NativeParser()
    return parser.parse_file(Path("test_dict_dict"))


def test_init() -> None:
    test_dict: SDict[str, Any] = SDict()
    assert test_dict.source_file is None
    assert test_dict.path == Path.cwd()
    assert test_dict.name == ""
    assert test_dict.line_content == []
    assert test_dict.line_comments == {}
    assert test_dict.includes == {}
    assert test_dict.block_content == ""
    assert test_dict.block_comments == {}
    assert test_dict.string_literals == {}
    assert test_dict.expressions == {}
    assert test_dict.delimiters == ["{", "}", "(", ")", "<", ">", ";", ","]


def test_init_with_file() -> None:
    test_dict: SDict[str, Any] = SDict("someDict")
    assert test_dict.source_file == Path.cwd() / "someDict"
    assert test_dict.path == Path.cwd()
    assert test_dict.name == "someDict"
    assert test_dict.line_content == []
    assert test_dict.line_comments == {}
    assert test_dict.includes == {}
    assert test_dict.block_content == ""
    assert test_dict.block_comments == {}
    assert test_dict.string_literals == {}
    assert test_dict.expressions == {}
    assert test_dict.delimiters == ["{", "}", "(", ")", "<", ">", ";", ","]


def test_init_with_base_dict() -> None:
    base_dict: dict[str, Any] = {
        "key1": "value1",
        "key2": "value2",
    }
    test_dict: SDict[str, Any] = SDict(base_dict)
    assert test_dict == base_dict
    assert test_dict.source_file is None
    assert test_dict.path == Path.cwd()
    assert test_dict.name == ""
    assert test_dict.line_content == []
    assert test_dict.line_comments == {}
    assert test_dict.includes == {}
    assert test_dict.block_content == ""
    assert test_dict.block_comments == {}
    assert test_dict.string_literals == {}
    assert test_dict.expressions == {}
    assert test_dict.delimiters == ["{", "}", "(", ")", "<", ">", ";", ","]


def test_find_global_key() -> None:
    # sourcery skip: no-loop-in-tests
    # Prepare
    str_in_1 = "PLACEHOLDER000001"
    str_in_2 = "PLACEHOLDER000002"
    str_in_3 = "PLACEHOLDER000003"
    not_a_str_1 = 1234
    not_a_str_2 = 1.23
    not_a_str_3 = False

    key_1 = "key_1"
    key_2 = "key_2"
    key_3 = "key_3"
    key_n_1 = "key_n_1"
    key_n_2 = "key_n_2"
    key_n_3 = "key_n_3"
    keyd = "keyd"
    keyl = "keyl"
    keydl = "keydl"
    keyld = "keyld"
    keydld = "keydld"
    keyldl = "keyldl"
    d_nested = {
        key_1: str_in_1,
        key_2: str_in_2,
        key_3: str_in_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
        key_n_3: not_a_str_3,
    }
    l_nested = [
        str_in_1,
        str_in_2,
        str_in_3,
        not_a_str_1,
        not_a_str_2,
        not_a_str_3,
    ]
    dl_nested = {keyl: deepcopy(l_nested)}
    ld_nested = [deepcopy(d_nested)]
    dld_nested = {keyld: deepcopy(ld_nested)}
    ldl_nested = [deepcopy(dl_nested)]

    # Construct a dictionary dict_in with single entries, nested dicts and nested lists
    dict_in: dict[str, Any] = {
        key_1: str_in_1,
        key_2: str_in_2,
        key_3: str_in_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
        key_n_3: not_a_str_3,
        keyd: d_nested,
        keyl: l_nested,
        keydl: dl_nested,
        keyld: ld_nested,
        keydld: dld_nested,
        keyldl: ldl_nested,
    }

    # Assert structure of dict_in
    assert len(dict_in) == 12
    assert isinstance(dict_in[keyd], dict)
    assert len(dict_in[keyd]) == 6
    assert len(dict_in[keyl]) == 6
    assert len(dict_in[keydl]) == 1
    assert len(dict_in[keydl][keyl]) == 6
    assert len(dict_in[keyld]) == 1
    assert len(dict_in[keyld][0]) == 6
    assert len(dict_in[keydld]) == 1
    assert len(dict_in[keydld][keyld]) == 1
    assert len(dict_in[keydld][keyld][0]) == 6
    assert len(dict_in[keyldl]) == 1
    assert len(dict_in[keyldl][0]) == 1
    assert len(dict_in[keyldl][0][keyl]) == 6

    # Assert content of dict_in
    assert dict_in[key_1] == str_in_1
    assert dict_in[key_2] == str_in_2
    assert dict_in[key_3] == str_in_3
    assert dict_in[keyd][key_1] == str_in_1
    assert dict_in[keyd][key_2] == str_in_2
    assert dict_in[keyd][key_3] == str_in_3
    assert dict_in[keyl][0] == str_in_1
    assert dict_in[keyl][1] == str_in_2
    assert dict_in[keyl][2] == str_in_3
    assert dict_in[keydl][keyl][0] == str_in_1
    assert dict_in[keydl][keyl][1] == str_in_2
    assert dict_in[keydl][keyl][2] == str_in_3
    assert dict_in[keyld][0][key_1] == str_in_1
    assert dict_in[keyld][0][key_2] == str_in_2
    assert dict_in[keyld][0][key_3] == str_in_3
    assert dict_in[keydld][keyld][0][key_1] == str_in_1
    assert dict_in[keydld][keyld][0][key_2] == str_in_2
    assert dict_in[keydld][keyld][0][key_3] == str_in_3
    assert dict_in[keyldl][0][keyl][0] == str_in_1
    assert dict_in[keyldl][0][keyl][1] == str_in_2
    assert dict_in[keyldl][0][keyl][2] == str_in_3

    # Execute find_global_key and set_global_key
    dict_out = deepcopy(dict_in)
    str_out_1 = "string 1"
    str_out_2 = "string 2"
    str_out_3 = "string 3"
    global_key = find_global_key(arg=dict_out, query=str_in_1)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_1 with str_out_1
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_1)
        global_key = find_global_key(arg=dict_out, query=str_in_1)
    assert i == 7
    global_key = find_global_key(arg=dict_out, query=str_in_2)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_2 with str_out_2
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_2)
        global_key = find_global_key(arg=dict_out, query=str_in_2)
    assert i == 7
    global_key = find_global_key(arg=dict_out, query=str_in_3)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_3 with str_out_3
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_3)
        global_key = find_global_key(arg=dict_out, query=str_in_3)
    assert i == 7

    # Assert structure of dict_out
    assert len(dict_out) == 12
    assert len(dict_out[keyd]) == 6
    assert len(dict_out[keyl]) == 6
    assert len(dict_out[keydl]) == 1
    assert len(dict_out[keydl][keyl]) == 6
    assert len(dict_out[keyld]) == 1
    assert len(dict_out[keyld][0]) == 6
    assert len(dict_out[keydld]) == 1
    assert len(dict_out[keydld][keyld]) == 1
    assert len(dict_out[keydld][keyld][0]) == 6
    assert len(dict_out[keyldl]) == 1
    assert len(dict_out[keyldl][0]) == 1
    assert len(dict_out[keyldl][0][keyl]) == 6

    # Assert content of dict_out
    assert dict_out[key_1] == str_out_1
    assert dict_out[key_2] == str_out_2
    assert dict_out[key_3] == str_out_3
    assert dict_out[keyd][key_1] == str_out_1
    assert dict_out[keyd][key_2] == str_out_2
    assert dict_out[keyd][key_3] == str_out_3
    assert dict_out[keyl][0] == str_out_1
    assert dict_out[keyl][1] == str_out_2
    assert dict_out[keyl][2] == str_out_3
    assert dict_out[keydl][keyl][0] == str_out_1
    assert dict_out[keydl][keyl][1] == str_out_2
    assert dict_out[keydl][keyl][2] == str_out_3
    assert dict_out[keyld][0][key_1] == str_out_1
    assert dict_out[keyld][0][key_2] == str_out_2
    assert dict_out[keyld][0][key_3] == str_out_3
    assert dict_out[keydld][keyld][0][key_1] == str_out_1
    assert dict_out[keydld][keyld][0][key_2] == str_out_2
    assert dict_out[keydld][keyld][0][key_3] == str_out_3
    assert dict_out[keyldl][0][keyl][0] == str_out_1
    assert dict_out[keyldl][0][keyl][1] == str_out_2
    assert dict_out[keyldl][0][keyl][2] == str_out_3


def test_order_keys() -> None:
    # sourcery skip: no-loop-in-tests
    # Prepare
    str_1 = "string 1"
    str_2 = "string 2"
    str_3 = "string 3"
    not_a_str_1 = 1234
    not_a_str_2 = 1.23
    not_a_str_3 = False

    key_1 = "key_1"
    key_2 = "key_2"
    key_3 = "key_3"
    key_n_1 = "key_n_1"
    key_n_2 = "key_n_2"
    key_n_3 = "key_n_3"
    key_4 = "key_4"

    d_nested = {
        key_3: str_3,
        key_1: str_1,
        key_2: str_2,
        key_n_3: not_a_str_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
    }
    dict_in: dict[str, Any] = {
        key_3: str_3,
        key_1: str_1,
        key_2: str_2,
        key_n_3: not_a_str_3,
        key_n_2: not_a_str_2,
        key_4: d_nested,
        key_n_1: not_a_str_1,
    }
    keys_expected = [key_1, key_2, key_3, key_4, key_n_1, key_n_2, key_n_3]
    keys_expected_nested = [key_1, key_2, key_3, key_n_1, key_n_2, key_n_3]

    s_dict: SDict[str, Any] = SDict()
    s_dict.update(deepcopy(dict_in))

    # 1. negative test: assert dict_in is not alphanumerically ordered
    for index, key in enumerate(dict_in):
        assert key != keys_expected[index]
    for index, key in enumerate(dict_in[key_4]):
        assert key != keys_expected_nested[index]

    # 2. negative test: assert dict is not alphanumerically ordered
    for index, key in enumerate(s_dict):
        assert key != keys_expected[index]
    for index, key in enumerate(s_dict[key_4]):
        assert key != keys_expected_nested[index]

    dict_out = order_keys(dict_in)  # order_keys function defined in dict.py module
    _s_dict_data_instance = s_dict
    s_dict.order_keys()  # order_keys instance method of SDict class

    # assert that instances dict_in and s_dict have been modified in place
    assert dict_out is dict_in
    assert s_dict is _s_dict_data_instance

    # 1. positive test for dict_out: assert dict_out is alphanumerically ordered
    for index, key in enumerate(dict_out):
        assert key == keys_expected[index]
    for index, key in enumerate(dict_out[key_4]):
        assert key == keys_expected_nested[index]

    # 2. positive test for dict: assert dict is alphanumerically ordered
    for index, key in enumerate(s_dict):
        assert key == keys_expected[index]
    for index, key in enumerate(s_dict[key_4]):
        assert key == keys_expected_nested[index]


def test_order_keys_of_test_dict(test_dict: SDict[K, V]) -> None:
    # Prepare
    # Execute
    test_dict.order_keys()
    # Assert
    assert str(test_dict[cast(K, "unordered")]) == str(test_dict[cast(K, "ordered")])


def test_reduce_scope_of_test_dict(test_dict: SDict[K, V]) -> None:
    # Prepare
    scope: list[K] = [cast(K, "scope"), cast(K, "subscope1")]
    # Execute
    test_dict.reduce_scope(scope)
    # Assert
    dict_out = test_dict
    assert len(dict_out) == 2  # subscope11, subscope12
    # assert dict_out["subscope11"] is not None
    assert dict_out[cast(K, "subscope11")]["name"] == "subscope11"
    assert dict_out[cast(K, "subscope12")] is not None
    assert dict_out[cast(K, "subscope12")]["name"] == "subscope12"


def test_include() -> None:
    # Prepare
    source_dict_file = Path("test_dict_add_include")
    param_dict_file = Path("test_dict_paramDict")
    temp_dict_file = Path(f"{source_dict_file.name}_temp")
    parsed_dict_file = Path(f"parsed.{temp_dict_file.name}")
    temp_dict_file.unlink(missing_ok=True)
    parsed_dict_file.unlink(missing_ok=True)
    source_dict = DictReader.read(source_dict_file)
    param_dict = DictReader.read(param_dict_file)
    # Execute
    source_dict.include(param_dict)
    # Assert
    DictWriter.write(source_dict, temp_dict_file)
    parsed_dict = DictParser.parse(temp_dict_file)
    assert parsed_dict is not None
    assert parsed_dict["valueA"] == 1
    # Clean up
    temp_dict_file.unlink(missing_ok=True)
    parsed_dict_file.unlink(missing_ok=True)


def test_merge_does_not_overwrite_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "B": 21,
            "C": 21.0,
            "D": True,
            "E": {
                "A": "string 22",
                "B": 22,
                "C": 22.0,
                "D": False,
            },
            "F": [
                "string 23",
                23,
                23.0,
                True,
            ],
        }
    )
    dict_1_original = deepcopy(dict_1)
    # merge dict_2 into dict_1
    dict_1.merge(dict_2)
    # assert that no entry in dict_1 has been overwritten
    # (because dict_2 contains only keys that are already present in dict_1)
    assert dict_1 == dict_1_original


def test_merge_does_not_delete_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "E": {},
            "F": [],
        }
    )
    dict_1_original = deepcopy(dict_1)
    # merge dict_2 into dict_1
    dict_1.merge(dict_2)
    # assert that no entry in dict_1 has been deleted
    assert dict_1 == dict_1_original


def test_merge_does_add_new_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "E": {
                "E": "string 24",
            },
            "G": "string 26",
        }
    )
    # merge dict_2 into dict_1
    dict_1.merge(dict_2)
    # assert dict_1 contains the new keys and values from dict_2
    assert dict_1["A"] == "string 11"
    assert dict_1["B"] == 11
    assert dict_1["C"] == 11.0
    assert dict_1["D"] is False
    assert dict_1["E"]["A"] == "string 12"
    assert dict_1["E"]["B"] == 12
    assert dict_1["E"]["C"] == 12.0
    assert dict_1["E"]["D"] is True
    assert dict_1["E"]["E"] == "string 24"
    assert dict_1["F"][0] == "string 13"
    assert dict_1["F"][1] == 13
    assert dict_1["F"][2] == 13.0
    assert dict_1["F"][3] is False
    assert dict_1["G"] == "string 26"


def test_merge_does_not_change_existings_lists() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "F": [
                "string 23",
                23,
                23.0,
                True,
                "string 25",
            ],
        }
    )
    # merge dict_2 into dict_1
    dict_1.merge(dict_2)
    # assert that list "F" remained unchanged and
    # that the additional element in list "F" has NOT been added to the list.
    assert dict_1["F"][0] == "string 13"
    assert dict_1["F"][1] == 13
    assert dict_1["F"][2] == 13.0
    assert dict_1["F"][3] is False
    assert len(dict_1["F"]) == 4
    with pytest.raises(IndexError):
        _ = dict_1["F"][4]


def test_merge_does_also_merge_attributes() -> None:
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_1.expressions |= {
        1: {
            "name": "EXPRESSION000011",
            "expression": "$varName11",
        },
        2: {
            "name": "EXPRESSION000012",
            "expression": "$varName12",
        },
    }
    dict_1.line_comments |= {
        1: "// line comment 11",
        2: "// line comment 12",
    }
    dict_1.block_comments |= {
        1: "/* block comment 11 */",
        2: "/* block comment 12 */",
    }
    dict_1.includes |= {
        1: ("#include dict_11", "dict_11", Path("dict_11")),
        2: ("#include dict_12", "dict_12", Path("dict_12")),
    }
    dict_2.expressions |= {
        1: {
            "name": "EXPRESSION000021",
            "expression": "$varName21",
        },
        2: {
            "name": "EXPRESSION000022",
            "expression": "$varName22",
        },
        3: {
            "name": "EXPRESSION000023",
            "expression": "$varName23",
        },
    }
    dict_2.line_comments |= {
        1: "// line comment 21",
        2: "// line comment 22",
        3: "// line comment 23",
    }
    dict_2.block_comments |= {
        1: "/* block comment 21 */",
        2: "/* block comment 22 */",
        3: "/* block comment 23 */",
    }
    dict_2.includes |= {
        1: ("#include dict_21", "dict_21", Path("dict_21")),
        2: ("#include dict_22", "dict_22", Path("dict_22")),
        3: ("#include dict_23", "dict_23", Path("dict_23")),
    }
    # merge dict_2 into dict_1
    dict_1.merge(dict_2)
    # assert that existing entries in the attributes of dict_1 have NOT been overwritten
    # and new entries have been added
    assert dict_1.expressions[1] == {
        "name": "EXPRESSION000011",
        "expression": "$varName11",
    }
    assert dict_1.expressions[2] == {
        "name": "EXPRESSION000012",
        "expression": "$varName12",
    }
    assert dict_1.expressions[3] == {
        "name": "EXPRESSION000023",
        "expression": "$varName23",
    }
    assert dict_1.line_comments[1] == "// line comment 11"
    assert dict_1.line_comments[2] == "// line comment 12"
    assert dict_1.line_comments[3] == "// line comment 23"
    assert dict_1.block_comments[1] == "/* block comment 11 */"
    assert dict_1.block_comments[2] == "/* block comment 12 */"
    assert dict_1.block_comments[3] == "/* block comment 23 */"
    assert dict_1.includes[1] == ("#include dict_11", "dict_11", Path("dict_11"))
    assert dict_1.includes[2] == ("#include dict_12", "dict_12", Path("dict_12"))
    assert dict_1.includes[3] == ("#include dict_23", "dict_23", Path("dict_23"))


def test_update_does_overwrite_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "B": 21,
            "C": 21.0,
            "D": True,
            "E": {
                "A": "string 22",
                "B": 22,
                "C": 22.0,
                "D": False,
            },
            "F": [
                "string 23",
                23,
                23.0,
                True,
            ],
        }
    )
    # update dict_1 with dict_2
    dict_1.update(dict_2)
    # assert all elements in dict_1 have been overwritten with the elements from dict_2
    assert dict_1["A"] == "string 21"
    assert dict_1["B"] == 21
    assert dict_1["C"] == 21.0
    assert dict_1["D"] is True
    assert dict_1["E"]["A"] == "string 22"
    assert dict_1["E"]["B"] == 22
    assert dict_1["E"]["C"] == 22.0
    assert dict_1["E"]["D"] is False
    assert dict_1["F"][0] == "string 23"
    assert dict_1["F"][1] == 23
    assert dict_1["F"][2] == 23.0
    assert dict_1["F"][3] is True


def test_update_does_delete_nested_elements() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "E": {},
            "F": [],
        }
    )
    # update dict_1 with dict_2
    dict_1.update(dict_2)
    # assert that formerly existing nested elements have been deleted,
    # because they are not present in dict_2
    assert dict_1["A"] == "string 21"  # overwritten by dict_2
    assert dict_1["B"] == 11  # not deleted by dict_2
    assert dict_1["C"] == 11.0  # not deleted by dict_2
    assert dict_1["D"] is False  # not deleted by dict_2
    assert dict_1["E"] == {}  # overwritten by dict_2, and hence all nested elements have been deleted
    assert dict_1["F"] == []  # overwritten by dict_2, and hence all nested elements have been deleted


def test_update_does_add_new_keys_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "E": {
                "E": "string 24",
            },
            "G": "string 26",
        }
    )
    # update dict_1 with dict_2
    dict_1.update(dict_2)
    # assert dict_1 contains the new keys, but nested keys which do not exist in dict_2 have been deleted
    assert dict_1["A"] == "string 11"
    assert dict_1["B"] == 11
    assert dict_1["C"] == 11.0
    assert dict_1["D"] is False
    with pytest.raises(KeyError):
        _ = dict_1["E"]["A"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["B"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["C"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["D"]
    assert dict_1["E"]["E"] == "string 24"
    assert dict_1["F"][0] == "string 13"
    assert dict_1["F"][1] == 13
    assert dict_1["F"][2] == 13.0
    assert dict_1["F"][3] is False
    assert dict_1["G"] == "string 26"


def test_update_does_change_existings_lists_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "F": [
                "string 23",
                23,
                23.0,
                True,
                "string 25",
            ],
        }
    )
    # update dict_1 with dict_2
    dict_1.update(dict_2)
    # assert that list "F" is overwritten with list "F" from dict_2
    assert dict_1["F"][0] == "string 23"
    assert dict_1["F"][1] == 23
    assert dict_1["F"][2] == 23.0
    assert dict_1["F"][3] is True
    assert dict_1["F"][4] == "string 25"


def test_update_does_also_update_attributes() -> None:
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_1.expressions |= {
        1: {
            "name": "EXPRESSION000011",
            "expression": "$varName11",
        },
        2: {
            "name": "EXPRESSION000012",
            "expression": "$varName12",
        },
    }
    dict_1.line_comments |= {
        1: "// line comment 11",
        2: "// line comment 12",
    }
    dict_1.block_comments |= {
        1: "/* block comment 11 */",
        2: "/* block comment 12 */",
    }
    dict_1.includes |= {
        1: ("#include dict_11", "dict_11", Path("dict_11")),
        2: ("#include dict_12", "dict_12", Path("dict_12")),
    }
    dict_2.expressions |= {
        1: {
            "name": "EXPRESSION000021",
            "expression": "$varName21",
        },
        2: {
            "name": "EXPRESSION000022",
            "expression": "$varName22",
        },
        3: {
            "name": "EXPRESSION000023",
            "expression": "$varName23",
        },
    }
    dict_2.line_comments |= {
        1: "// line comment 21",
        2: "// line comment 22",
        3: "// line comment 23",
    }
    dict_2.block_comments |= {
        1: "/* block comment 21 */",
        2: "/* block comment 22 */",
        3: "/* block comment 23 */",
    }
    dict_2.includes |= {
        1: ("#include dict_21", "dict_21", Path("dict_21")),
        2: ("#include dict_22", "dict_22", Path("dict_22")),
        3: ("#include dict_23", "dict_23", Path("dict_23")),
    }
    # update dict_1 with dict_2
    dict_1.update(dict_2)
    # assert that existing entries in the attributes of dict_1 HAVE been overwritten
    # and new entries have been added
    assert dict_1.expressions[1] == {
        "name": "EXPRESSION000021",
        "expression": "$varName21",
    }
    assert dict_1.expressions[2] == {
        "name": "EXPRESSION000022",
        "expression": "$varName22",
    }
    assert dict_1.expressions[3] == {
        "name": "EXPRESSION000023",
        "expression": "$varName23",
    }
    assert dict_1.line_comments[1] == "// line comment 21"
    assert dict_1.line_comments[2] == "// line comment 22"
    assert dict_1.line_comments[3] == "// line comment 23"
    assert dict_1.block_comments[1] == "/* block comment 21 */"
    assert dict_1.block_comments[2] == "/* block comment 22 */"
    assert dict_1.block_comments[3] == "/* block comment 23 */"
    assert dict_1.includes[1] == ("#include dict_21", "dict_21", Path("dict_21"))
    assert dict_1.includes[2] == ("#include dict_22", "dict_22", Path("dict_22"))
    assert dict_1.includes[3] == ("#include dict_23", "dict_23", Path("dict_23"))


def test_augmented_or_does_overwrite_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "B": 21,
            "C": 21.0,
            "D": True,
            "E": {
                "A": "string 22",
                "B": 22,
                "C": 22.0,
                "D": False,
            },
            "F": [
                "string 23",
                23,
                23.0,
                True,
            ],
        }
    )
    # execute augmented or operation
    dict_1 |= dict_2
    # assert all elements in dict_1 have been overwritten with the elements from dict_2
    assert dict_1["A"] == "string 21"
    assert dict_1["B"] == 21
    assert dict_1["C"] == 21.0
    assert dict_1["D"] is True
    assert dict_1["E"]["A"] == "string 22"
    assert dict_1["E"]["B"] == 22
    assert dict_1["E"]["C"] == 22.0
    assert dict_1["E"]["D"] is False
    assert dict_1["F"][0] == "string 23"
    assert dict_1["F"][1] == 23
    assert dict_1["F"][2] == 23.0
    assert dict_1["F"][3] is True


def test_augmented_or_does_delete_nested_elements() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "E": {},
            "F": [],
        }
    )
    # execute augmented or operation
    dict_1 |= dict_2
    # assert that formerly existing nested elements have been deleted,
    # because they are not present in dict_2
    assert dict_1["A"] == "string 21"  # overwritten by dict_2
    assert dict_1["B"] == 11  # not deleted by dict_2
    assert dict_1["C"] == 11.0  # not deleted by dict_2
    assert dict_1["D"] is False  # not deleted by dict_2
    assert dict_1["E"] == {}  # overwritten by dict_2, and hence all nested elements have been deleted
    assert dict_1["F"] == []  # overwritten by dict_2, and hence all nested elements have been deleted


def test_augmented_or_does_add_new_keys_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "E": {
                "E": "string 24",
            },
            "G": "string 26",
        }
    )
    # execute augmented or operation
    dict_1 |= dict_2
    # assert dict_1 contains the new keys, but nested keys which do not exist in dict_2 have been deleted
    assert dict_1["A"] == "string 11"
    assert dict_1["B"] == 11
    assert dict_1["C"] == 11.0
    assert dict_1["D"] is False
    with pytest.raises(KeyError):
        _ = dict_1["E"]["A"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["B"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["C"]
    with pytest.raises(KeyError):
        _ = dict_1["E"]["D"]
    assert dict_1["E"]["E"] == "string 24"
    assert dict_1["F"][0] == "string 13"
    assert dict_1["F"][1] == 13
    assert dict_1["F"][2] == 13.0
    assert dict_1["F"][3] is False
    assert dict_1["G"] == "string 26"


def test_augmented_or_does_change_existings_lists_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "F": [
                "string 23",
                23,
                23.0,
                True,
                "string 25",
            ],
        }
    )
    # execute augmented or operation
    dict_1 |= dict_2
    # assert that list "F" is overwritten with list "F" from dict_2
    assert dict_1["F"][0] == "string 23"
    assert dict_1["F"][1] == 23
    assert dict_1["F"][2] == 23.0
    assert dict_1["F"][3] is True
    assert dict_1["F"][4] == "string 25"


def test_augmented_or_does_also_update_attributes() -> None:
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_1.expressions |= {
        1: {
            "name": "EXPRESSION000011",
            "expression": "$varName11",
        },
        2: {
            "name": "EXPRESSION000012",
            "expression": "$varName12",
        },
    }
    dict_1.line_comments |= {
        1: "// line comment 11",
        2: "// line comment 12",
    }
    dict_1.block_comments |= {
        1: "/* block comment 11 */",
        2: "/* block comment 12 */",
    }
    dict_1.includes |= {
        1: ("#include dict_11", "dict_11", Path("dict_11")),
        2: ("#include dict_12", "dict_12", Path("dict_12")),
    }
    dict_2.expressions |= {
        1: {
            "name": "EXPRESSION000021",
            "expression": "$varName21",
        },
        2: {
            "name": "EXPRESSION000022",
            "expression": "$varName22",
        },
        3: {
            "name": "EXPRESSION000023",
            "expression": "$varName23",
        },
    }
    dict_2.line_comments |= {
        1: "// line comment 21",
        2: "// line comment 22",
        3: "// line comment 23",
    }
    dict_2.block_comments |= {
        1: "/* block comment 21 */",
        2: "/* block comment 22 */",
        3: "/* block comment 23 */",
    }
    dict_2.includes |= {
        1: ("#include dict_21", "dict_21", Path("dict_21")),
        2: ("#include dict_22", "dict_22", Path("dict_22")),
        3: ("#include dict_23", "dict_23", Path("dict_23")),
    }
    # execute augmented or operation
    dict_1 |= dict_2
    # assert that existing entries in the attributes of dict_1 HAVE been overwritten
    # and new entries have been added
    assert dict_1.expressions[1] == {
        "name": "EXPRESSION000021",
        "expression": "$varName21",
    }
    assert dict_1.expressions[2] == {
        "name": "EXPRESSION000022",
        "expression": "$varName22",
    }
    assert dict_1.expressions[3] == {
        "name": "EXPRESSION000023",
        "expression": "$varName23",
    }
    assert dict_1.line_comments[1] == "// line comment 21"
    assert dict_1.line_comments[2] == "// line comment 22"
    assert dict_1.line_comments[3] == "// line comment 23"
    assert dict_1.block_comments[1] == "/* block comment 21 */"
    assert dict_1.block_comments[2] == "/* block comment 22 */"
    assert dict_1.block_comments[3] == "/* block comment 23 */"
    assert dict_1.includes[1] == ("#include dict_21", "dict_21", Path("dict_21"))
    assert dict_1.includes[2] == ("#include dict_22", "dict_22", Path("dict_22"))
    assert dict_1.includes[3] == ("#include dict_23", "dict_23", Path("dict_23"))


def test_left_or_does_overwrite_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "B": 21,
            "C": 21.0,
            "D": True,
            "E": {
                "A": "string 22",
                "B": 22,
                "C": 22.0,
                "D": False,
            },
            "F": [
                "string 23",
                23,
                23.0,
                True,
            ],
        }
    )
    # execute left or operation
    new_dict = dict_1 | dict_2
    # assert all elements in dict_1 have been overwritten with the elements from dict_2
    assert new_dict["A"] == "string 21"
    assert new_dict["B"] == 21
    assert new_dict["C"] == 21.0
    assert new_dict["D"] is True
    assert new_dict["E"]["A"] == "string 22"
    assert new_dict["E"]["B"] == 22
    assert new_dict["E"]["C"] == 22.0
    assert new_dict["E"]["D"] is False
    assert new_dict["F"][0] == "string 23"
    assert new_dict["F"][1] == 23
    assert new_dict["F"][2] == 23.0
    assert new_dict["F"][3] is True


def test_left_or_does_delete_nested_elements() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "E": {},
            "F": [],
        }
    )
    # execute left or operation
    new_dict = dict_1 | dict_2
    # assert that formerly existing nested elements have been deleted,
    # because they are not present in dict_2
    assert new_dict["A"] == "string 21"  # overwritten by dict_2
    assert new_dict["B"] == 11  # not deleted by dict_2
    assert new_dict["C"] == 11.0  # not deleted by dict_2
    assert new_dict["D"] is False  # not deleted by dict_2
    assert new_dict["E"] == {}  # overwritten by dict_2, and hence all nested elements have been deleted
    assert new_dict["F"] == []  # overwritten by dict_2, and hence all nested elements have been deleted


def test_left_or_does_add_new_keys_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "E": {
                "E": "string 24",
            },
            "G": "string 26",
        }
    )
    # execute left or operation
    new_dict = dict_1 | dict_2
    # assert dict_1 contains the new keys, but nested keys which do not exist in dict_2 have been deleted
    assert new_dict["A"] == "string 11"
    assert new_dict["B"] == 11
    assert new_dict["C"] == 11.0
    assert new_dict["D"] is False
    with pytest.raises(KeyError):
        _ = new_dict["E"]["A"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["B"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["C"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["D"]
    assert new_dict["E"]["E"] == "string 24"
    assert new_dict["F"][0] == "string 13"
    assert new_dict["F"][1] == 13
    assert new_dict["F"][2] == 13.0
    assert new_dict["F"][3] is False
    assert new_dict["G"] == "string 26"


def test_left_or_does_change_existings_lists_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "F": [
                "string 23",
                23,
                23.0,
                True,
                "string 25",
            ],
        }
    )
    # execute left or operation
    new_dict = dict_1 | dict_2
    # assert that list "F" is overwritten with list "F" from dict_2
    assert new_dict["F"][0] == "string 23"
    assert new_dict["F"][1] == 23
    assert new_dict["F"][2] == 23.0
    assert new_dict["F"][3] is True
    assert new_dict["F"][4] == "string 25"


def test_left_or_does_also_update_attributes() -> None:
    dict_1: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_1.expressions |= {
        1: {
            "name": "EXPRESSION000011",
            "expression": "$varName11",
        },
        2: {
            "name": "EXPRESSION000012",
            "expression": "$varName12",
        },
    }
    dict_1.line_comments |= {
        1: "// line comment 11",
        2: "// line comment 12",
    }
    dict_1.block_comments |= {
        1: "/* block comment 11 */",
        2: "/* block comment 12 */",
    }
    dict_1.includes |= {
        1: ("#include dict_11", "dict_11", Path("dict_11")),
        2: ("#include dict_12", "dict_12", Path("dict_12")),
    }
    dict_2.expressions |= {
        1: {
            "name": "EXPRESSION000021",
            "expression": "$varName21",
        },
        2: {
            "name": "EXPRESSION000022",
            "expression": "$varName22",
        },
        3: {
            "name": "EXPRESSION000023",
            "expression": "$varName23",
        },
    }
    dict_2.line_comments |= {
        1: "// line comment 21",
        2: "// line comment 22",
        3: "// line comment 23",
    }
    dict_2.block_comments |= {
        1: "/* block comment 21 */",
        2: "/* block comment 22 */",
        3: "/* block comment 23 */",
    }
    dict_2.includes |= {
        1: ("#include dict_21", "dict_21", Path("dict_21")),
        2: ("#include dict_22", "dict_22", Path("dict_22")),
        3: ("#include dict_23", "dict_23", Path("dict_23")),
    }
    # execute left or operation
    new_dict: SDict[str, Any | dict[str | int, Any]]
    new_dict = dict_1 | dict_2
    # assert that existing entries in the attributes of dict_1 HAVE been overwritten
    # and new entries have been added
    assert new_dict.expressions[1] == {
        "name": "EXPRESSION000021",
        "expression": "$varName21",
    }
    assert new_dict.expressions[2] == {
        "name": "EXPRESSION000022",
        "expression": "$varName22",
    }
    assert new_dict.expressions[3] == {
        "name": "EXPRESSION000023",
        "expression": "$varName23",
    }
    assert new_dict.line_comments[1] == "// line comment 21"
    assert new_dict.line_comments[2] == "// line comment 22"
    assert new_dict.line_comments[3] == "// line comment 23"
    assert new_dict.block_comments[1] == "/* block comment 21 */"
    assert new_dict.block_comments[2] == "/* block comment 22 */"
    assert new_dict.block_comments[3] == "/* block comment 23 */"
    assert new_dict.includes[1] == ("#include dict_21", "dict_21", Path("dict_21"))
    assert new_dict.includes[2] == ("#include dict_22", "dict_22", Path("dict_22"))
    assert new_dict.includes[3] == ("#include dict_23", "dict_23", Path("dict_23"))


class DictWithoutOr(dict[Any, Any]):
    def __or__(self, value: MutableMapping[Any, Any]) -> dict[Any, Any]:
        return NotImplemented


def test_right_or_does_overwrite_existing_keys() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: dict[str, Any | dict[str | int, Any]] = DictWithoutOr(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "B": 21,
            "C": 21.0,
            "D": True,
            "E": {
                "A": "string 22",
                "B": 22,
                "C": 22.0,
                "D": False,
            },
            "F": [
                "string 23",
                23,
                23.0,
                True,
            ],
        }
    )
    # execute right or operation
    new_dict = dict_1 | dict_2
    # assert all elements in dict_1 have been overwritten with the elements from dict_2
    assert new_dict["A"] == "string 21"
    assert new_dict["B"] == 21
    assert new_dict["C"] == 21.0
    assert new_dict["D"] is True
    assert new_dict["E"]["A"] == "string 22"
    assert new_dict["E"]["B"] == 22
    assert new_dict["E"]["C"] == 22.0
    assert new_dict["E"]["D"] is False
    assert new_dict["F"][0] == "string 23"
    assert new_dict["F"][1] == 23
    assert new_dict["F"][2] == 23.0
    assert new_dict["F"][3] is True


def test_right_or_does_delete_nested_elements() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: dict[str, Any | dict[str | int, Any]] = DictWithoutOr(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "A": "string 21",
            "E": {},
            "F": [],
        }
    )
    # execute right or operation
    new_dict = dict_1 | dict_2
    # assert that formerly existing nested elements have been deleted,
    # because they are not present in dict_2
    assert new_dict["A"] == "string 21"  # overwritten by dict_2
    assert new_dict["B"] == 11  # not deleted by dict_2
    assert new_dict["C"] == 11.0  # not deleted by dict_2
    assert new_dict["D"] is False  # not deleted by dict_2
    assert new_dict["E"] == {}  # overwritten by dict_2, and hence all nested elements have been deleted
    assert new_dict["F"] == []  # overwritten by dict_2, and hence all nested elements have been deleted


def test_right_or_does_add_new_keys_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: dict[str, Any | dict[str | int, Any]] = DictWithoutOr(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "E": {
                "E": "string 24",
            },
            "G": "string 26",
        }
    )
    # execute right or operation
    new_dict = dict_1 | dict_2
    # assert dict_1 contains the new keys, but nested keys which do not exist in dict_2 have been deleted
    assert new_dict["A"] == "string 11"
    assert new_dict["B"] == 11
    assert new_dict["C"] == 11.0
    assert new_dict["D"] is False
    with pytest.raises(KeyError):
        _ = new_dict["E"]["A"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["B"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["C"]
    with pytest.raises(KeyError):
        _ = new_dict["E"]["D"]
    assert new_dict["E"]["E"] == "string 24"
    assert new_dict["F"][0] == "string 13"
    assert new_dict["F"][1] == 13
    assert new_dict["F"][2] == 13.0
    assert new_dict["F"][3] is False
    assert new_dict["G"] == "string 26"


def test_right_or_does_change_existings_lists_by_overwrite() -> None:
    # construct two dicts with single entries, a nested dict and a nested list
    dict_1: dict[str, Any | dict[str | int, Any]] = DictWithoutOr(
        {
            "A": "string 11",
            "B": 11,
            "C": 11.0,
            "D": False,
            "E": {
                "A": "string 12",
                "B": 12,
                "C": 12.0,
                "D": True,
            },
            "F": [
                "string 13",
                13,
                13.0,
                False,
            ],
        }
    )
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict(
        {
            "F": [
                "string 23",
                23,
                23.0,
                True,
                "string 25",
            ],
        }
    )
    # execute right or operation
    new_dict = dict_1 | dict_2
    # assert that list "F" is overwritten with list "F" from dict_2
    assert new_dict["F"][0] == "string 23"
    assert new_dict["F"][1] == 23
    assert new_dict["F"][2] == 23.0
    assert new_dict["F"][3] is True
    assert new_dict["F"][4] == "string 25"


def test_right_or_does_also_update_attributes() -> None:
    dict_1: dict[str, Any | dict[str | int, Any]] = DictWithoutOr()
    dict_2: SDict[str, Any | dict[str | int, Any]] = SDict()
    dict_2.expressions |= {
        1: {
            "name": "EXPRESSION000021",
            "expression": "$varName21",
        },
        2: {
            "name": "EXPRESSION000022",
            "expression": "$varName22",
        },
        3: {
            "name": "EXPRESSION000023",
            "expression": "$varName23",
        },
    }
    dict_2.line_comments |= {
        1: "// line comment 21",
        2: "// line comment 22",
        3: "// line comment 23",
    }
    dict_2.block_comments |= {
        1: "/* block comment 21 */",
        2: "/* block comment 22 */",
        3: "/* block comment 23 */",
    }
    dict_2.includes |= {
        1: ("#include dict_21", "dict_21", Path("dict_21")),
        2: ("#include dict_22", "dict_22", Path("dict_22")),
        3: ("#include dict_23", "dict_23", Path("dict_23")),
    }
    # execute right or operation
    new_dict: SDict[str, Any | dict[str | int, Any]]
    new_dict = dict_1 | dict_2  # type: ignore[assignment, reportAssignmentType]
    # assert that existing entries in the attributes of dict_1 HAVE been overwritten
    # and new entries have been added
    assert new_dict.expressions[1] == {
        "name": "EXPRESSION000021",
        "expression": "$varName21",
    }
    assert new_dict.expressions[2] == {
        "name": "EXPRESSION000022",
        "expression": "$varName22",
    }
    assert new_dict.expressions[3] == {
        "name": "EXPRESSION000023",
        "expression": "$varName23",
    }
    assert new_dict.line_comments[1] == "// line comment 21"
    assert new_dict.line_comments[2] == "// line comment 22"
    assert new_dict.line_comments[3] == "// line comment 23"
    assert new_dict.block_comments[1] == "/* block comment 21 */"
    assert new_dict.block_comments[2] == "/* block comment 22 */"
    assert new_dict.block_comments[3] == "/* block comment 23 */"
    assert new_dict.includes[1] == ("#include dict_21", "dict_21", Path("dict_21"))
    assert new_dict.includes[2] == ("#include dict_22", "dict_22", Path("dict_22"))
    assert new_dict.includes[3] == ("#include dict_23", "dict_23", Path("dict_23"))


def test_dict_copy() -> None:
    original_dict = _construct_test_dict()
    manual_copy = _construct_test_dict()
    # execute copy operation
    copied_dict = original_dict.copy()
    # assert that the copied dict is of type dict
    assert isinstance(copied_dict, dict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a shallow copy: the nested dict and list are the same objects
    assert copied_dict["E"] is original_dict["E"]
    assert copied_dict["F"] is original_dict["F"]
    # Changing an element in the nested dict should change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 22"
    # Changing an element in the nested list should change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 23"
    # However, changing the nested dict or list itself should not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 22",  # (had been changed from "string 12" to "string 22" above)
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 23",  # (had been changed from "string 13" to "string 23" above)
        13,
        13.0,
        False,
    ]
    # assert that the attributes are the same objects


def test_dict_copy_copy() -> None:
    original_dict = _construct_test_dict()
    manual_copy = _construct_test_dict()
    # execute copy operation
    copied_dict = copy.copy(original_dict)
    # assert that the copied dict is of type dict
    assert isinstance(copied_dict, dict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a shallow copy: the nested dict and list are the same objects
    assert copied_dict["E"] is original_dict["E"]
    assert copied_dict["F"] is original_dict["F"]
    # Changing an element in the nested dict should change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 22"
    # Changing an element in the nested list should change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 23"
    # However, changing the nested dict or list itself should not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 22",  # (had been changed from "string 12" to "string 22" above)
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 23",  # (had been changed from "string 13" to "string 23" above)
        13,
        13.0,
        False,
    ]
    # assert that the attributes are the same objects


def test_dict_copy_deepcopy() -> None:
    original_dict = _construct_test_dict()
    manual_copy = _construct_test_dict()
    # execute deepcopy operation
    copied_dict = copy.deepcopy(original_dict)
    # assert that the copied dict is of type dict
    assert isinstance(copied_dict, dict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a deep copy: even the nested dict and list are not the same objects
    assert copied_dict["E"] is not original_dict["E"]
    assert copied_dict["F"] is not original_dict["F"]
    # They contain equal elements, though
    assert copied_dict["E"] == original_dict["E"]
    assert copied_dict["F"] == original_dict["F"]
    # Changing an element in the nested dict in the copy does not change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 12"  # (unchanged)
    # Changing an element in the nested list in the copy does not change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 13"  # (unchanged)
    # Likewise, changing the nested dict or list completely does not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 12",
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 13",
        13,
        13.0,
        False,
    ]
    # assert that the attributes are NOT the same objects
    # They contain equal elements, though


def test_sdict_copy() -> None:
    original_dict = _construct_test_sdict()
    manual_copy = _construct_test_sdict()
    # execute copy operation
    copied_dict = original_dict.copy()
    # assert that the copied dict is of type SDict
    assert isinstance(copied_dict, SDict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a shallow copy: the nested dict and list are the same objects
    assert copied_dict["E"] is original_dict["E"]
    assert copied_dict["F"] is original_dict["F"]
    # Changing an element in the nested dict should change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 22"
    # Changing an element in the nested list should change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 23"
    # However, changing the nested dict or list itself should not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 22",  # (had been changed from "string 12" to "string 22" above)
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 23",  # (had been changed from "string 13" to "string 23" above)
        13,
        13.0,
        False,
    ]
    # assert that the attributes are the same objects
    assert copied_dict.expressions is original_dict.expressions
    assert copied_dict.line_comments is original_dict.line_comments
    assert copied_dict.block_comments is original_dict.block_comments
    assert copied_dict.includes is original_dict.includes


def test_sdict_copy_copy() -> None:
    original_dict = _construct_test_sdict()
    manual_copy = _construct_test_sdict()
    # execute copy operation
    copied_dict = copy.copy(original_dict)
    # assert that the copied dict is of type SDict
    assert isinstance(copied_dict, SDict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a shallow copy: the nested dict and list are the same objects
    assert copied_dict["E"] is original_dict["E"]
    assert copied_dict["F"] is original_dict["F"]
    # Changing an element in the nested dict should change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 22"
    # Changing an element in the nested list should change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 23"
    # However, changing the nested dict or list itself should not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 22",  # (had been changed from "string 12" to "string 22" above)
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 23",  # (had been changed from "string 13" to "string 23" above)
        13,
        13.0,
        False,
    ]
    # assert that the attributes are the same objects
    assert copied_dict.expressions is original_dict.expressions
    assert copied_dict.line_comments is original_dict.line_comments
    assert copied_dict.block_comments is original_dict.block_comments
    assert copied_dict.includes is original_dict.includes


def test_sdict_copy_deepcopy() -> None:
    original_dict = _construct_test_sdict()
    manual_copy = _construct_test_sdict()
    # execute deepcopy operation
    copied_dict = copy.deepcopy(original_dict)
    # assert that the copied dict is of type SDict
    assert isinstance(copied_dict, SDict)
    # assert that the copied dict is equal to the manually constructed copy
    assert copied_dict == manual_copy
    # assert that the copied dict is not the same object as the original dict
    assert copied_dict is not original_dict
    # assert that the first level elements are not the same, but got copied:
    # Changing the copied dict should not change the original dict
    copied_dict["A"] = "string 21"
    assert copied_dict["A"] == "string 21"
    assert original_dict["A"] == "string 11"
    # assert that the copy is a deep copy: even the nested dict and list are not the same objects
    assert copied_dict["E"] is not original_dict["E"]
    assert copied_dict["F"] is not original_dict["F"]
    # They contain equal elements, though
    assert copied_dict["E"] == original_dict["E"]
    assert copied_dict["F"] == original_dict["F"]
    # Changing an element in the nested dict in the copy does not change the original dict
    copied_dict["E"]["A"] = "string 22"
    assert copied_dict["E"]["A"] == "string 22"
    assert original_dict["E"]["A"] == "string 12"  # (unchanged)
    # Changing an element in the nested list in the copy does not change the original list
    copied_dict["F"][0] = "string 23"
    assert copied_dict["F"][0] == "string 23"
    assert original_dict["F"][0] == "string 13"  # (unchanged)
    # Likewise, changing the nested dict or list completely does not change the original dict
    copied_dict["E"] = {}
    assert copied_dict["E"] == {}
    assert original_dict["E"] == {
        "A": "string 12",
        "B": 12,
        "C": 12.0,
        "D": True,
    }
    copied_dict["F"] = []
    assert copied_dict["F"] == []
    assert original_dict["F"] == [
        "string 13",
        13,
        13.0,
        False,
    ]
    # assert that the attributes are NOT the same objects
    assert copied_dict.expressions is not original_dict.expressions
    assert copied_dict.line_comments is not original_dict.line_comments
    assert copied_dict.block_comments is not original_dict.block_comments
    assert copied_dict.includes is not original_dict.includes
    # They contain equal elements, though
    assert copied_dict.expressions == original_dict.expressions
    assert copied_dict.line_comments == original_dict.line_comments
    assert copied_dict.block_comments == original_dict.block_comments
    assert copied_dict.includes == original_dict.includes


def _construct_test_dict() -> dict[str, Any]:
    # construct a test dict with single entries, a nested dict and a nested list
    test_dict: dict[str, Any] = {
        "A": "string 11",
        "B": 11,
        "C": 11.0,
        "D": False,
        "E": {
            "A": "string 12",
            "B": 12,
            "C": 12.0,
            "D": True,
        },
        "F": [
            "string 13",
            13,
            13.0,
            False,
        ],
    }
    return test_dict


def _construct_test_sdict() -> SDict[str, Any]:
    # construct a test SDict with single entries, a nested dict and a nested list
    test_sdict: SDict[str, Any] = SDict(_construct_test_dict())
    test_sdict.expressions |= {
        1: {
            "name": "EXPRESSION000011",
            "expression": "$varName11",
        },
        2: {
            "name": "EXPRESSION000012",
            "expression": "$varName12",
        },
    }
    test_sdict.line_comments |= {
        1: "// line comment 11",
        2: "// line comment 12",
    }
    test_sdict.block_comments |= {
        1: "/* block comment 11 */",
        2: "/* block comment 12 */",
    }
    test_sdict.includes |= {
        1: ("#include dict_11", "dict_11", Path("dict_11")),
        2: ("#include dict_12", "dict_12", Path("dict_12")),
    }
    return test_sdict


def test_load() -> None:
    # Prepare
    source_file = Path("test_dictReader_dict")
    s_dict: SDict[str, Any] = SDict()
    # Execute
    _ = s_dict.load(source_file)
    # Assert included dict has been merged
    assert s_dict["paramA"] == 3.0
    assert s_dict["paramB"] == 4.0
    assert s_dict["paramC"] == 7.0
    assert s_dict["paramD"] == 0.66
    assert s_dict["paramE"] == [0.1, 0.2, 0.4]
    assert s_dict["paramF"] == [[0.3, 0.9], [2.7, 8.1]]
    assert s_dict["paramG"] == [[10, "fancy", 3.14, "s"], ["more", 2, "come"]]
    # Assert references and expressions have been resolved
    assert s_dict["keyA"] == 3.0
    assert s_dict["keyB"] == 4.0
    assert s_dict["keyC"] == 7.0
    assert s_dict["keyD"] == 4.16
    assert s_dict["keyE"] == 6.67
    assert s_dict["keyF"] == 8.98
    assert s_dict["keyG"] == 20.34
    assert s_dict["keyH"] == 0.2
    assert s_dict["keyI"] == [0.1, 0.2, 0.4]
    assert s_dict["keyJ"] == 8.1
    assert s_dict["keyK"] == [2.7, 8.1]
    assert s_dict["keyL"] == [[0.3, 0.9], [2.7, 8.1]]
    assert s_dict["keyM"] == 9.0
    assert s_dict["keyN"] == 7.0


def test_dump() -> None:
    # Prepare
    target_file: Path = Path("temp_file_test_write_dict")
    test_dict: dict[str, Any] = {
        "param1": -10.0,
        "param2": 0.0,
        "param3": 0.0,
        "_case": {
            "_layer": "lhsvar",
            "_level": 1,
            "_no_of_samples": 108,
            "_index": 0,
            "_path": r"C:\Users\CLAROS\Documents\SystemSimulation\ModelVerification\tools\farn\test\cases\gp_0\lhsvar_000",  # noqa: E501
            "_key": "lhsvar_000",
            "_is_leaf": False,
            "_condition": None,
            "_names": ["param1", "param2", "param3"],
            "_values": [-10.0, 0.0, 0.0],
            "_commands": {"ls": ["echo %PATH%", "dir"]},
        },
    }
    test_str: str = (
        r"/*---------------------------------*- C++ -*----------------------------------*\ "
        r"filetype dictionary; coding utf-8; version 0.1; local --; purpose --; "
        r"\*----------------------------------------------------------------------------*/ "
        r"param1 -10.0; param2 0.0; param3 0.0; "
        r"_case { _layer lhsvar; _level 1; _no_of_samples 108; _index 0; "
        r"_path 'C:\Users\CLAROS\Documents\SystemSimulation\ModelVerification\tools\farn\test\cases\gp_0\lhsvar_000'; "
        r"_key lhsvar_000; _is_leaf false; _condition NULL; "
        r"_names ( param1 param2 param3 ); _values ( -10.0 0.0 0.0 ); "
        r"_commands { ls ( 'echo %PATH%' dir ); } } "
    )
    test_s_dict: SDict[str, Any]
    # Execute 1: Dump with explicit target_file
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict()
    test_s_dict.update(test_dict)
    _ = test_s_dict.dump(target_file)
    # Assert 1
    assert target_file.exists()
    parsed_str = re.sub(r"[\r\n\s]+", " ", str(DictReader.read(target_file)))
    assert parsed_str == test_str
    assert test_s_dict.source_file is not None
    assert test_s_dict.source_file.exists()
    assert test_s_dict.source_file == target_file.absolute()
    assert test_s_dict.path == target_file.absolute().parent
    assert test_s_dict.name == target_file.absolute().name

    # Execute 2: Dump with explicit target_file, while source_file is set to same file as target_file
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict(target_file)
    test_s_dict.update(test_dict)
    _ = test_s_dict.dump(target_file)
    # Assert 2
    assert target_file.exists()
    parsed_str = re.sub(r"[\r\n\s]+", " ", str(DictReader.read(target_file)))
    assert parsed_str == test_str
    assert test_s_dict.source_file is not None
    assert test_s_dict.source_file.exists()
    assert test_s_dict.source_file == target_file.absolute()
    assert test_s_dict.path == target_file.absolute().parent
    assert test_s_dict.name == target_file.absolute().name

    # Execute 3: Dump with explicit target_file, while source_file is different than target_file
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict("some_source_file")
    test_s_dict.update(test_dict)
    _ = test_s_dict.dump(target_file)
    # Assert 3
    assert target_file.exists()
    parsed_str = re.sub(r"[\r\n\s]+", " ", str(DictReader.read(target_file)))
    assert parsed_str == test_str
    assert test_s_dict.source_file is not None
    assert test_s_dict.source_file.exists()
    assert test_s_dict.source_file == target_file.absolute()
    assert test_s_dict.path == target_file.absolute().parent
    assert test_s_dict.name == target_file.absolute().name

    # Execute 4: Dump without passing target_file, while source_file is set to target_file
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict(target_file)
    test_s_dict.update(test_dict)
    _ = test_s_dict.dump()
    # Assert 4
    assert target_file.exists()
    parsed_str = re.sub(r"[\r\n\s]+", " ", str(DictReader.read(target_file)))
    assert parsed_str == test_str
    assert test_s_dict.source_file is not None
    assert test_s_dict.source_file.exists()
    assert test_s_dict.source_file == target_file.absolute()
    assert test_s_dict.path == target_file.absolute().parent
    assert test_s_dict.name == target_file.absolute().name

    # Execute 5: Dump without passing target_file, while source_file is set some other source file
    _source_file_name = "some_source_file"
    _source_file = Path(_source_file_name)
    _source_file.unlink(missing_ok=True)
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict(_source_file_name)
    test_s_dict.update(test_dict)
    _ = test_s_dict.dump()
    # Assert 5
    assert _source_file.exists()
    assert not target_file.exists()
    parsed_str = re.sub(r"[\r\n\s]+", " ", str(DictReader.read(_source_file)))
    assert parsed_str == test_str
    assert test_s_dict.source_file is not None
    assert test_s_dict.source_file.exists()
    assert test_s_dict.source_file == _source_file.absolute()
    assert test_s_dict.path == _source_file.absolute().parent
    assert test_s_dict.name == _source_file.absolute().name
    _source_file.unlink(missing_ok=True)

    # Execute 6: Dump without passing target_file, while source_file is NOT set
    target_file.unlink(missing_ok=True)
    test_s_dict = SDict()
    test_s_dict.update(test_dict)
    # Assert 6
    with pytest.raises(ValueError):
        _ = test_s_dict.dump()

    # Clean up
    _source_file.unlink(missing_ok=True)
    target_file.unlink(missing_ok=True)


def test_dump_load_example_used_in_docs() -> None:
    # Prepare
    my_dict: SDict[str, int] = SDict(
        {
            "foo": 1,
            "bar": 2,
        }
    )
    # Execute
    target_file = my_dict.dump("myDict")
    my_dict_loaded: SDict[str, int] = SDict().load("myDict")
    # Assert
    assert my_dict_loaded == my_dict
    # Clean up
    target_file.unlink(missing_ok=True)


def test_permissable_key_type_example_used_in_docs() -> None:
    # Prepare
    str_dict: SDict[str, Any] = SDict(
        {
            "foo": 1,
            "bar": 2,
        }
    )
    int_dict: SDict[int, Any] = SDict(
        {
            1: "foo",
            2: "bar",
        }
    )
    str_or_int_dict: SDict[str | int, Any] = SDict(
        {
            "foo": 1,
            2: "bar",
        }
    )
    # Execute
    str_dict_file = str_dict.dump("temp_str_dict")
    int_dict_file = int_dict.dump("temp_int_Dict")
    str_or_int_dict_file = str_or_int_dict.dump("temp_str_or_int_dict")
    str_dict_loaded: SDict[str, Any] = SDict().load(str_dict_file)
    int_dict_loaded: SDict[int, Any] = SDict().load(int_dict_file)
    str_or_int_dict_loaded: SDict[str | int, Any] = SDict().load(str_or_int_dict_file)
    # Assert
    assert str_dict_loaded == str_dict
    assert int_dict_loaded == int_dict
    assert str_or_int_dict_loaded == str_or_int_dict
    # Clean up
    str_dict_file.unlink(missing_ok=True)
    int_dict_file.unlink(missing_ok=True)
    str_or_int_dict_file.unlink(missing_ok=True)


def test_cpp_dict() -> None:
    s_dict: SDict[str, Any] = SDict(_construct_test_dict())
    cpp_dict: CppDict = CppDict(_construct_test_dict())
    assert isinstance(cpp_dict, dict)
    assert isinstance(cpp_dict, SDict)
    assert isinstance(cpp_dict, CppDict)
    assert cpp_dict == s_dict



import re
from copy import copy, deepcopy
from pathlib import Path
from typing import Any

import pytest

from dictIO import DictReader, FoamFormatter, Formatter, NativeFormatter, SDict, XmlFormatter


class TestFormatter:
    # @TODO: To be implemented
    @pytest.mark.skip(reason="To be implemented")
    def test_returned_formatter_type(self) -> None:
        pass

    def test_format_value_bool(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = Formatter()
        bool_in = False
        str_out = formatter.format_value(bool_in)
        assert isinstance(str_out, str)
        assert str_out == "False"
        bool_in = True
        str_out = formatter.format_value(bool_in)
        assert isinstance(str_out, str)
        assert str_out == "True"


class TestNativeFormatter:
    @pytest.mark.parametrize(
        "str_in",
        [
            "string",
            "0.1",
            "2",
            "+0.1" "+2",
            "-0.1",
            "-2",
            "$keyword",
            "$keyword1",
            "$keyword1[0]",
            "$keyword1[1][2]",
        ],
    )
    def test_format_value_string_no_additional_quotes_expected(self, str_in: str) -> None:
        # Prepare
        formatter = NativeFormatter()
        str_expected = str_in
        # Execute
        str_out = formatter.format_value(str_in)
        # Assert
        assert str_out == str_expected

    @pytest.mark.parametrize(
        "str_in",
        [
            "a string with spaces",
            r"C:\a\path\in\windows",
            r"C:/a/path/in/linux",
            "",
            r'contains a "nested string" literal with double quotes',
            r"contains a \"nested string\" literal with escaped double quotes",
        ],
    )
    def test_format_value_string_additional_single_quotes_expected(self, str_in: str) -> None:
        # Prepare
        formatter = NativeFormatter()
        str_expected = f"'{str_in}'"
        # Execute
        str_out = formatter.format_value(str_in)
        # Assert
        assert str_out == str_expected

    @pytest.mark.parametrize(
        "str_in",
        [
            "$keyword+1",
            "$keyword - 3.0",
            "$keyword1 * $keyword2",
            r"contains a 'nested string' literal with single quotes",
            r"contains a \'nested string\' literal with escaped single quotes",
        ],
    )
    def test_format_value_string_additional_double_quotes_expected(self, str_in: str) -> None:
        # Prepare
        formatter = NativeFormatter()
        str_expected = f'"{str_in}"'
        # Execute
        str_out = formatter.format_value(str_in)
        # Assert
        assert str_out == str_expected

    def test_format_value_float(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = NativeFormatter()
        float_in = 1.23
        str_out = formatter.format_value(float_in)
        assert isinstance(str_out, str)
        assert str_out == "1.23"
        float_in = 1.0
        str_out = formatter.format_value(float_in)
        assert isinstance(str_out, str)
        assert str_out == "1.0"
        float_in = 0.0
        str_out = formatter.format_value(float_in)
        assert isinstance(str_out, str)
        assert str_out == "0.0"

    def test_format_value_int(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = NativeFormatter()
        int_in = 1234
        str_out = formatter.format_value(int_in)
        assert isinstance(str_out, str)
        assert str_out == "1234"
        int_in = -1234
        str_out = formatter.format_value(int_in)
        assert isinstance(str_out, str)
        assert str_out == "-1234"
        int_in = 0
        str_out = formatter.format_value(int_in)
        assert isinstance(str_out, str)
        assert str_out == "0"
        int_in = 1
        str_out = formatter.format_value(int_in)
        assert isinstance(str_out, str)
        assert str_out == "1"

    def test_format_value_bool(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = NativeFormatter()
        bool_in = False
        str_out = formatter.format_value(bool_in)
        assert isinstance(str_out, str)
        assert str_out == "false"
        bool_in = True
        str_out = formatter.format_value(bool_in)
        assert isinstance(str_out, str)
        assert str_out == "true"

    def test_format_value_Path(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = NativeFormatter()
        path_in = Path("folder1/folder2/file.txt")
        str_out = formatter.format_value(path_in)
        assert isinstance(str_out, str)
        assert str_out == str(path_in)

    def test_format_value_None(self) -> None:
        # sourcery skip: extract-duplicate-method, inline-variable
        formatter = NativeFormatter()
        none_in = None
        str_out = formatter.format_value(none_in)
        assert isinstance(str_out, str)
        assert str_out == "NULL"

    def test_insert_block_comments(self) -> None:
        # sourcery skip: class-extract-method
        # Prepare
        formatter = NativeFormatter()
        as_is_block_comment = (
            "/*---------------------------------*- C++ -*----------------------------------*\\\n"
            "This is a block comment; coding utf-8; version 0.1;\n"
            "\\*----------------------------------------------------------------------------*/"
        )
        default_block_comment = (
            "/*---------------------------------*- C++ -*----------------------------------*\\\n"
            "filetype dictionary; coding utf-8; version 0.1; local --; purpose --;\n"
            "\\*----------------------------------------------------------------------------*/"
        )
        # Execute and Assert
        # (dispatch to run_block_comment_tests())
        TestNativeFormatter.run_block_comment_tests(
            formatter,
            as_is_block_comment,
            default_block_comment,
        )

    @staticmethod
    def run_block_comment_tests(
        formatter: NativeFormatter,
        as_is_block_comment: str,
        default_block_comment: str,
    ) -> None:
        s_dict: SDict[str, Any] = SDict()
        # as we used test_simpleDict, str_in does not have a block comment yet
        str_in_template = formatter.format_dict(s_dict)
        placeholder1 = "BLOCKCOMMENT000101            BLOCKCOMMENT000101;"
        placeholder2 = "BLOCKCOMMENT000102            BLOCKCOMMENT000102;"
        placeholder3 = "BLOCKCOMMENT000103            BLOCKCOMMENT000103;"

        # STANDARD CASE: The dictionary contains 1 (ONE) BLOCK COMMENT
        # Prepare
        s_dict.block_comments = {101: as_is_block_comment}
        str_in = placeholder1 + "\n" + str_in_template
        str_expected = str_in.replace(placeholder1, as_is_block_comment)
        # Execute
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

        # FALLBACK CASE: The dictionary contains 0 (NO) BLOCK COMMENT
        # Prepare
        s_dict.block_comments = {}
        str_in = str_in_template
        str_expected = default_block_comment + "\n" + str_in
        # Execute
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected
        # Execute a second time:
        #   Does it still work when we call insert_block_comments() the second time?
        #   Will the default block comment then still be inserted?
        str_in = str_in_template
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

        # NON-STANDARD CASE 1: The dictionary contains 3 (THREE) BLOCK COMMENTS, ALL IDENTICAL
        # Prepare
        s_dict.block_comments = {
            101: as_is_block_comment,
            102: as_is_block_comment,
            103: as_is_block_comment,
        }
        str_in = placeholder1 + "\n" + placeholder2 + "\n" + placeholder3 + "\n" + str_in_template
        str_expected = (
            str_in.replace(placeholder1, as_is_block_comment).replace(placeholder2, "").replace(placeholder3, "")
        )
        # Execute
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

        # NON-STANDARD CASE 2: The dictionary contains 3 (THREE) BLOCK COMMENTS, NON IDENTICAL
        # Prepare
        s_dict.block_comments = {
            101: as_is_block_comment,
            102: default_block_comment,
            103: as_is_block_comment,
        }
        str_in = placeholder1 + "\n" + placeholder2 + "\n" + placeholder3 + "\n" + str_in_template
        str_expected = (
            str_in.replace(placeholder1, as_is_block_comment)
            .replace(placeholder2, default_block_comment)
            .replace(placeholder3, "")
        )
        # Execute
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

        # NON-STANDARD CASE 3: The dictionary contains 1 (ONE) BLOCK COMMENT, BUT IT DOES NOT CONTAIN ' C++ '
        # Prepare
        block_comment_tampered = re.sub(r"\s[Cc]\+{2}\s", " C# ", as_is_block_comment)
        s_dict.block_comments = {101: block_comment_tampered}
        str_in = placeholder1 + "\n" + str_in_template
        str_expected = str_in.replace(placeholder1, default_block_comment + "\n" + block_comment_tampered)
        # Execute
        str_out = formatter.insert_block_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

    def test_insert_includes(self) -> None:
        # Prepare
        include_directive_in = '#include "test formatter paramDict"'
        include_file_name_in = "test formatter paramDict"
        include_file_path_in = Path("test formatter paramDict").absolute()
        s_dict: SDict[str, Any] = SDict()
        s_dict.includes[102] = (
            include_directive_in,
            include_file_name_in,
            include_file_path_in,
        )
        blockcomment_placeholder = "BLOCKCOMMENT000101            BLOCKCOMMENT000101;"
        include_placeholder = "INCLUDE000102            INCLUDE000102;"
        str_in = blockcomment_placeholder + "\n" + include_placeholder + "\n"
        str_expected = str_in.replace(include_placeholder, include_directive_in.replace('"', "'"))
        formatter = NativeFormatter()
        # Execute
        str_out = formatter.insert_includes(s_dict, str_in)
        # Assert
        assert str_out == str_expected

    def test_insert_line_comments(self) -> None:
        # Prepare
        s_dict: SDict[str, Any] = SDict()
        line_comment_in = "// This is a line comment"
        formatter = NativeFormatter()
        str_in_template = formatter.format_dict(s_dict)
        s_dict.line_comments = {103: line_comment_in}
        placeholder1 = "BLOCKCOMMENT000101            BLOCKCOMMENT000101;"
        placeholder2 = "INCLUDE000102            INCLUDE000102;"
        placeholder3 = "LINECOMMENT000103            LINECOMMENT000103;"
        str_in = placeholder1 + "\n" + placeholder2 + "\n" + placeholder3 + "\n" + str_in_template
        str_expected = str_in.replace(placeholder3, line_comment_in)
        # Execute
        str_out = formatter.insert_line_comments(s_dict, str_in)
        # Assert
        assert str_out == str_expected

    def test_remove_trailing_spaces(self) -> None:
        # sourcery skip: extract-duplicate-method, move-assign-in-block
        # Prepare
        # Construct 7 test strings, each having a different number of leading and trailing spaces
        str_in_1 = "a string with one trailing space "
        str_expected_1 = "a string with one trailing space"
        str_in_2 = "a string with five trailing spaces     "
        str_expected_2 = "a string with five trailing spaces"
        str_in_3 = " a string with one leading space"
        str_expected_3 = " a string with one leading space"
        str_in_4 = "     a string with five leading spaces"
        str_expected_4 = "     a string with five leading spaces"
        str_in_5 = " a string with one leading and one trailing space "
        str_expected_5 = " a string with one leading and one trailing space"
        str_in_6 = "     a string with five leading and five trailing spaces     "
        str_expected_6 = "     a string with five leading and five trailing spaces"
        str_in_7 = "a string with no leading or trailing spaces"
        str_expected_7 = "a string with no leading or trailing spaces"

        multi_line_str_in = ""
        multi_line_str_in += str_in_1 + "\n"
        multi_line_str_in += str_in_2 + "\n"
        multi_line_str_in += str_in_3 + "\n"
        multi_line_str_in += str_in_4 + "\n"
        multi_line_str_in += str_in_5 + "\n"
        multi_line_str_in += str_in_6 + "\n"
        multi_line_str_in += str_in_7

        multi_line_str_expected = ""
        multi_line_str_expected += str_expected_1 + "\n"
        multi_line_str_expected += str_expected_2 + "\n"
        multi_line_str_expected += str_expected_3 + "\n"
        multi_line_str_expected += str_expected_4 + "\n"
        multi_line_str_expected += str_expected_5 + "\n"
        multi_line_str_expected += str_expected_6 + "\n"
        multi_line_str_expected += str_expected_7

        formatter = NativeFormatter()

        # Execute 1
        multi_line_str_out = formatter.remove_trailing_spaces(multi_line_str_in)
        # Assert 1
        assert multi_line_str_out == multi_line_str_expected

        # Execute 2: Same test, but this time with the last line having a line ending
        multi_line_str_in += "\n"
        multi_line_str_expected += "\n"
        multi_line_str_out = formatter.remove_trailing_spaces(multi_line_str_in)
        # Assert 2
        assert multi_line_str_out == multi_line_str_expected

        # Execute 3: Only one single line
        str_out = formatter.remove_trailing_spaces(str_in_1)
        # Assert 3
        assert str_out == str_expected_1

        # Execute 4: Empty string
        str_in = ""
        str_out = formatter.remove_trailing_spaces(str_in)
        # Assert 4
        assert str_out == ""

    def test_list_with_nested_list(self) -> None:
        # Prepare
        test_obj: dict[str, Any] = {
            "blocks": [
                "hex",
                [0, 1, 2, 3, 4, 5, 6, 7],
                [1, 1, 2],
                "simpleGrading",
                [1, 1, 1],
                "hex",
                [10, 11, 12, 13, 14, 15, 16, 17],
                [1, 1, 2],
                "edgeGrading",
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            ]
        }
        s_dict: SDict[str, Any] = SDict()
        dict_in = deepcopy(s_dict)
        dict_in.update(test_obj)
        formatter = NativeFormatter()
        assert test_obj == dict_in
        # Execute
        str_in: str = formatter.format_dict(test_obj)
        str_out: str = formatter.format_dict(dict_in)
        # Assert
        assert str_in == str_out

    def test_to_string_does_not_alter_original(self) -> None:
        # Prepare
        dict_in = DictReader.read(Path("test_formatter_dict"))
        formatter = NativeFormatter()
        dict_in_reference = dict_in
        dict_in_shallowcopy = copy(dict_in)
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert isinstance(str_out, str)
        # Make sure that the original dictionary is not modified
        assert dict_in == dict_in_reference
        assert dict_in is dict_in_reference
        for key in dict_in:
            assert dict_in[key] == dict_in_reference[key]
            assert dict_in[key] is dict_in_reference[key]
        assert dict_in == dict_in_shallowcopy
        assert dict_in is not dict_in_shallowcopy
        for key in dict_in:
            assert dict_in[key] == dict_in_shallowcopy[key]
            assert dict_in[key] is dict_in_shallowcopy[key]


class TestFoamFormatter:
    def test_insert_block_comments(self) -> None:
        # sourcery skip: class-extract-method
        # Prepare
        formatter = FoamFormatter()
        as_is_block_comment = (
            "/*--------------------------------*- C++ -*----------------------------------*\\\n"
            "| =========       This is a block comment                                     |\n"
            "| \\\\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |\n"
            "|  \\\\    /   O peration     | Version:  dev                                   |\n"
            "|   \\\\  /    A nd           | Web:      www.OpenFOAM.com                      |\n"
            "|    \\\\/     M anipulation  |                                                 |\n"
            "\\*---------------------------------------------------------------------------*/\n"
            "FoamFile\n"
            "{\n"
            "    version                   2.0;\n"
            "    format                    ascii;\n"
            "    class                     dictionary;\n"
            "    object                    foamDict;\n"
            "}\n"
            "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //"
        )
        default_block_comment = (
            "/*--------------------------------*- C++ -*----------------------------------*\\\n"
            "| =========                 |                                                 |\n"
            "| \\\\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |\n"
            "|  \\\\    /   O peration     | Version:  dev                                   |\n"
            "|   \\\\  /    A nd           | Web:      www.OpenFOAM.com                      |\n"
            "|    \\\\/     M anipulation  |                                                 |\n"
            "\\*---------------------------------------------------------------------------*/\n"
            "FoamFile\n"
            "{\n"
            "    version                   2.0;\n"
            "    format                    ascii;\n"
            "    class                     dictionary;\n"
            "    object                    foamDict;\n"
            "}\n"
            "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //"
        )
        # Execute and Assert
        # (dispatch to TestNativeFormatter as, apart from the block comment itself, the tests are identical)
        TestNativeFormatter.run_block_comment_tests(
            formatter,
            as_is_block_comment,
            default_block_comment,
        )

    def test_insert_includes(self) -> None:
        # Prepare
        include_directive_in = "#include 'test formatter paramDict'"
        include_file_name_in = "test formatter paramDict"
        include_file_path_in = Path("test formatter paramDict").absolute()
        s_dict: SDict[str, Any] = SDict()
        s_dict.includes[102] = (
            include_directive_in,
            include_file_name_in,
            include_file_path_in,
        )
        blockcomment_placeholder = "BLOCKCOMMENT000101            BLOCKCOMMENT000101;"
        include_placeholder = "INCLUDE000102            INCLUDE000102;"
        str_in = blockcomment_placeholder + "\n" + include_placeholder + "\n"
        str_expected = str_in.replace(include_placeholder, include_directive_in.replace("'", '"'))
        formatter = FoamFormatter()
        # Execute
        str_out: str = formatter.insert_includes(s_dict, str_in)
        # Assert
        assert str_out == str_expected

    def test_ensure_string_does_not_contain_single_quotes(self) -> None:
        # Prepare dict until and including ()
        dict_in = DictReader.read(Path("test_formatter_dict"), comments=True)
        formatter = FoamFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert re.search(r"\'", str_out) is None

    def test_ensure_string_does_not_contain_underscore_variables(self) -> None:
        # Prepare dict until and including ()
        dict_in = DictReader.read(Path("test_formatter_dict"), comments=False)
        formatter = FoamFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert re.search(r"\s+_", str_out) is None

    def test_to_string_does_not_alter_original(self) -> None:
        # Prepare
        dict_in = DictReader.read(Path("test_formatter_dict"))
        formatter = FoamFormatter()
        dict_in_reference = dict_in
        dict_in_shallowcopy = copy(dict_in)
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert isinstance(str_out, str)
        # Make sure that the original dictionary is not modified
        assert dict_in == dict_in_reference
        assert dict_in is dict_in_reference
        for key in dict_in:
            assert dict_in[key] == dict_in_reference[key]
            assert dict_in[key] is dict_in_reference[key]
        assert dict_in == dict_in_shallowcopy
        assert dict_in is not dict_in_shallowcopy
        for key in dict_in:
            assert dict_in[key] == dict_in_shallowcopy[key]
            assert dict_in[key] is dict_in_shallowcopy[key]


class TestXmlFormatter:
    def test_default_options(self) -> None:
        # Execute
        formatter = XmlFormatter()
        # Assert
        assert formatter.omit_prefix is True
        assert formatter.integrate_attributes is True
        assert formatter.remove_node_numbering is True
        # assert formatter.indent == ' ' * 4
        # assert formatter.root_tag == 'NOTSPECIFIED'
        # assert formatter.namespaces == {
        #     'xs': 'https://www.w3.org/2009/XMLSchema/XMLSchema.xsd'
        # }
        # assert formatter.root_attributes is None
        # assert len(formatter.prefixes) == 1
        # assert formatter.prefixes[0] == 'xs'
        # assert formatter.attributes == {}
        # assert formatter.root_element.tag.find(
        #     'https://www.w3.org/2009/XMLSchema/XMLSchema.xsd'
        # ) > 0

    def test_to_string(self) -> None:
        # Prepare
        source_file = Path("test_formatter_dict")
        target_file = Path(f"parsed.{source_file}.xml")
        target_file.unlink(missing_ok=True)
        dict_in = DictReader.read(source_file)
        xml_opts = {
            "_nameSpaces": {"osp": "https://opensimulationplatform.com/xsd/OspModelDescription"},
            "_rootTag": "OspModelDescription",
            "_removeNodeNumbering": True,
        }
        dict_in.update({"_xmlOpts": xml_opts})
        formatter = XmlFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        # @TODO: Nothing is de facto asserted here. By intention?
        # -> Needs to checked and properly implemented. CLAROS, 2021-12-23
        with Path.open(target_file, "w") as f:
            _ = f.write(str_out)
        # Clean up
        target_file.unlink()

    def test_parse_format_reparse(self) -> None:
        # Prepare XML string to be parsed
        from dictIO import XmlParser

        file_name = Path("test_parser_xml.xml")
        str_in = ""
        with Path.open(file_name) as f:
            str_in = f.read()
        parser = XmlParser(add_node_numbering=False)
        formatter = XmlFormatter()
        dict_parsed: SDict[str, Any] = SDict()
        dict_reparsed: SDict[str, Any] = SDict()
        # Execute
        dict_parsed = parser.parse_string(str_in, dict_parsed)
        str_out: str = formatter.to_string(dict_parsed)
        dict_reparsed = parser.parse_string(str_out, dict_reparsed)
        # Assert
        assert dict_reparsed == dict_parsed

    def test_format_xml_namespace_explicit(self) -> None:
        # sourcery skip: class-extract-method
        # Prepare
        source_file = Path("test_formatter_dict")
        dict_in = DictReader.read(source_file)
        xml_opts = {
            "_nameSpaces": {"osp": "https://opensimulationplatform.com/xsd/OspModelDescription"},
            "_rootTag": "OspModelDescription",
            "_removeNodeNumbering": True,
        }
        dict_in.update({"_xmlOpts": xml_opts})
        formatter = XmlFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert 'xmlns:osp="https://opensimulationplatform.com/xsd/OspModelDescription"' in str_out
        assert 'xmlns="https://opensimulationplatform.com/xsd/OspModelDescription"' not in str_out

    def test_format_xml_namespace_default(self) -> None:
        # Prepare
        source_file = Path("test_formatter_dict")
        dict_in = DictReader.read(source_file)
        xml_opts = {
            "_nameSpaces": {"None": "https://opensimulationplatform.com/xsd/OspModelDescription"},
            "_rootTag": "OspModelDescription",
            "_removeNodeNumbering": True,
        }
        dict_in.update({"_xmlOpts": xml_opts})
        formatter = XmlFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert 'xmlns="https://opensimulationplatform.com/xsd/OspModelDescription"' in str_out
        assert 'xmlns:None="https://opensimulationplatform.com/xsd/OspModelDescription"' not in str_out

    @pytest.mark.skip(reason="XML pretty printing is not solved yet. The root attribute for encoding still gets lost.")
    def test_format_xml_root_attributes(self) -> None:
        # Prepare
        source_file = Path("test_formatter_dict")
        dict_in = DictReader.read(source_file)
        xml_opts = {
            "_nameSpaces": {"xs": "http://www.w3.org/2001/XMLSchema"},
            "_rootTag": "ROOT",
            "_rootAttributes": {
                "version": 0.1,
                "encoding": "UTF-8",
            },
            "_removeNodeNumbering": True,
        }
        dict_in.update({"_xmlOpts": xml_opts})
        formatter = XmlFormatter()
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert '?xml version="1.0" encoding="UTF-8"?' in str_out
        assert '?xml version="1.0"?' not in str_out

    def test_to_string_does_not_alter_original(self) -> None:
        # Prepare
        dict_in = DictReader.read(Path("test_formatter_dict"))
        formatter = XmlFormatter()
        dict_in_reference = dict_in
        dict_in_shallowcopy = copy(dict_in)
        # Execute
        str_out: str = formatter.to_string(dict_in)
        # Assert
        assert isinstance(str_out, str)
        # Make sure that the original dictionary is not modified
        assert dict_in == dict_in_reference
        assert dict_in is dict_in_reference
        for key in dict_in:
            assert dict_in[key] == dict_in_reference[key]
            assert dict_in[key] is dict_in_reference[key]
        assert dict_in == dict_in_shallowcopy
        assert dict_in is not dict_in_shallowcopy
        for key in dict_in:
            assert dict_in[key] == dict_in_shallowcopy[key]
            assert dict_in[key] is dict_in_shallowcopy[key]



import re
from copy import deepcopy
from pathlib import Path, PurePath
from typing import Any

import pytest

from dictIO import DictReader, DictWriter, FoamParser, SDict, create_target_file_name
from dictIO.utils.counter import BorgCounter


def test_write_dict() -> None:
    # sourcery skip: extract-duplicate-method
    # Prepare
    target_file: Path = Path("temp_file_test_write_dict")
    test_dict: dict[str, Any] = {
        "param1": -10.0,
        "param2": 0.0,
        "param3": 0.0,
        "_case": {
            "_layer": "lhsvar",
            "_level": 1,
            "_no_of_samples": 108,
            "_index": 0,
            "_path": r"C:\Users\CLAROS\Documents\SystemSimulation\ModelVerification\tools\farn\test\cases\gp_0\lhsvar_000",  # noqa: E501
            "_key": "lhsvar_000",
            "_is_leaf": False,
            "_condition": None,
            "_names": ["param1", "param2", "param3"],
            "_values": [-10.0, 0.0, 0.0],
            "_commands": {"ls": ["echo %PATH%", "dir"]},
        },
    }
    test_str: str = (
        r"/*---------------------------------*- C++ -*----------------------------------*\ "
        r"filetype dictionary; coding utf-8; version 0.1; local --; purpose --; "
        r"\*----------------------------------------------------------------------------*/ "
        r"param1 -10.0; param2 0.0; param3 0.0; "
        r"_case { _layer lhsvar; _level 1; _no_of_samples 108; _index 0; "
        r"_path 'C:\Users\CLAROS\Documents\SystemSimulation\ModelVerification\tools\farn\test\cases\gp_0\lhsvar_000'; "
        r"_key lhsvar_000; _is_leaf false; _condition NULL; "
        r"_names ( param1 param2 param3 ); _values ( -10.0 0.0 0.0 ); "
        r"_commands { ls ( 'echo %PATH%' dir ); } } "
    )
    test_s_dict: SDict[str, Any] = SDict()
    test_s_dict.update(test_dict)

    # Execute 1.1: Write as SDict completely new
    target_file.unlink(missing_ok=True)
    DictWriter.write(source_dict=test_s_dict, target_file=target_file, mode="w")
    # Assert 1.1
    assert target_file.exists()
    parsed_str = re.sub(
        pattern=r"[\r\n\s]+",
        repl=" ",
        string=str(DictReader.read(source_file=target_file)),
    )
    assert parsed_str == test_str

    # Execute 1.2: Write as SDict on top of existing
    DictWriter.write(test_s_dict, target_file, mode="a")
    # Assert 1.2
    assert target_file.exists()
    parsed_str = re.sub(
        pattern=r"[\r\n\s]+",
        repl=" ",
        string=str(DictReader.read(source_file=target_file)),
    )
    assert parsed_str == test_str

    # Execute 2.1: Write as dict completely new
    target_file.unlink()
    DictWriter.write(test_dict, target_file, mode="w")
    # Assert 2.1
    assert target_file.exists()
    parsed_str = re.sub(
        pattern=r"[\r\n\s]+",
        repl=" ",
        string=str(DictReader.read(source_file=target_file)),
    )
    assert parsed_str == test_str

    # Execute 2.2: Write as dict on top of existent
    DictWriter.write(test_dict, target_file, mode="a")
    # Assert 2.2
    assert target_file.exists()
    parsed_str = re.sub(
        pattern=r"[\r\n\s]+",
        repl=" ",
        string=str(DictReader.read(source_file=target_file)),
    )
    assert parsed_str == test_str

    # Clean up
    target_file.unlink()


def test_write_mode() -> None:
    # sourcery skip: extract-duplicate-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file}")
    target_file.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=False)
    DictWriter.write(dict_read, target_file)
    # delete one element (parameterD) in the dict to have a test case for the write mode
    dict_read_modified = deepcopy(dict_read)
    del dict_read_modified["parameters"]["parameterD"]

    # Execute mode 'a'
    # -> parameterD should still exist after writing (as the existing file was not overwritten but only appended to)
    DictWriter.write(dict_read_modified, target_file, mode="a")
    dict_reread = DictReader.read(target_file, includes=False)
    # Assert mode 'a'
    assert target_file.exists()
    assert "parameterD" in dict_reread["parameters"]

    # Execute mode 'w'
    # -> parameterD should NOT exist after writing (as the existing file was overwritten)
    DictWriter.write(dict_read_modified, target_file, mode="w")
    dict_reread = DictReader.read(target_file, includes=False)
    # Assert mode 'w'
    assert target_file.exists()
    assert "parameterD" not in dict_reread["parameters"]

    # Clean up
    target_file.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_dict(
    *,
    includes: bool,
) -> None:
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes)
    # Execute
    DictWriter.write(dict_read, target_file)
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


def test_read_dict_write_dict_target_file_given_as_str() -> None:
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path.cwd().absolute() / "parsed.test_dictWriter_dict"
    target_file_as_str = str(target_file.absolute())
    target_file.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file)
    # Execute
    DictWriter.write(dict_read, target_file_as_str)
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


def test_read_dict_write_dict_target_file_given_as_purepath() -> None:
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path.cwd().absolute() / "parsed.test_dictWriter_dict"
    target_file_as_purepath = PurePath(str(target_file.absolute()))
    target_file.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file)
    # Execute
    DictWriter.write(dict_read, target_file_as_purepath)
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_abc(
    *,
    includes: bool,
) -> None:
    # sourcery skip: class-extract-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file_abc = Path(f"{target_file.name}.abc")
    target_file.unlink(missing_ok=True)
    target_file_abc.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes)
    # Execute
    DictWriter.write(dict_read, target_file_abc)
    dict_reread = DictReader.read(target_file_abc)
    # Assert
    assert not target_file.exists()
    assert target_file_abc.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file_abc.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_cpp(
    *,
    includes: bool,
) -> None:
    # sourcery skip: class-extract-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file_cpp = Path(f"{target_file.name}.cpp")
    target_file.unlink(missing_ok=True)
    target_file_cpp.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes)
    # Execute
    DictWriter.write(dict_read, target_file_cpp)
    dict_reread = DictReader.read(target_file_cpp)
    # Assert
    assert not target_file.exists()
    assert target_file_cpp.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file_cpp.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_foam(
    *,
    includes: bool,
) -> None:
    # sourcery skip: class-extract-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file_foam = Path(f"{target_file.name}.foam")
    target_file.unlink(missing_ok=True)
    target_file_foam.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes, comments=False)
    # Execute
    DictWriter.write(dict_read, target_file_foam)
    dict_reread = DictReader.read(target_file_foam, includes=includes, comments=False)
    # Assert
    # (The 'FoamFile' element is foam dicts specific.
    #  Before comparison with the original cpp dict, we need to delete it.)
    del dict_reread["FoamFile"]
    assert not target_file.exists()
    assert target_file_foam.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file_foam.unlink()


# @TODO: The read_cpp_write_xml test fails as XmlParser and XmlFormatter
#        do not support yet all dict elements.
#        We might want to improve the XML parser and formatter
#        one day to the point these tests pass.
#        CLAROS, 2021-12-22
@pytest.mark.skip(
    reason=(
        "XmlParser and XmlFormatter do not support yet all dict elements.\n"
        "E.g. the XMLFormatter does currently not support writing None / NULL values. "
        "This causes this test to fail,\n"
        "as the None values in the source file (test_dictWriter_dict) "
        "are not persisted in XML and get hence lost on rereading."
    )
)
@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_xml(
    *,
    includes: bool,
) -> None:
    # sourcery skip: class-extract-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file_xml = Path(f"{target_file.name}.xml")
    target_file.unlink(missing_ok=True)
    target_file_xml.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes, comments=False)
    # Execute
    DictWriter.write(dict_read, target_file_xml)
    dict_reread = DictReader.read(target_file_xml)
    # Assert
    assert not target_file.exists()
    assert target_file_xml.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file_xml.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_dict_write_json(
    *,
    includes: bool,
) -> None:
    # sourcery skip: class-extract-method
    # Prepare
    source_file = Path("test_dictWriter_dict")
    target_file = Path(f"parsed.{source_file.name}")
    target_file_json = Path(f"{target_file.name}.json")
    target_file.unlink(missing_ok=True)
    target_file_json.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes, comments=False)
    # Execute
    DictWriter.write(dict_read, target_file_json)
    dict_reread = DictReader.read(target_file_json)
    # Assert
    assert not target_file.exists()
    assert target_file_json.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file_json.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_foam_write_foam(
    *,
    includes: bool,
) -> None:
    # Prepare
    source_file = Path("test_dictWriter_foam")
    target_file = Path(f"parsed.{source_file}.foam")
    target_file.unlink(missing_ok=True)
    parser = FoamParser()
    dict_read = DictReader.read(source_file, includes=includes, parser=parser)
    # Execute
    DictWriter.write(dict_read, target_file)
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_xml_write_xml(
    *,
    includes: bool,
) -> None:
    # Prepare
    source_file = Path("test_dictWriter_xml.xml")
    target_file = Path(f"parsed.{source_file}")
    target_file.unlink(missing_ok=True)
    BorgCounter.reset()
    dict_read = DictReader.read(source_file, includes=includes)
    # Execute
    DictWriter.write(dict_read, target_file)
    BorgCounter.reset()
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


@pytest.mark.parametrize("includes", [False, True])
def test_read_json_write_json(
    *,
    includes: bool,
) -> None:
    # Prepare
    source_file = Path("test_dictWriter_json.json")
    target_file = Path(f"parsed.{source_file}")
    target_file.unlink(missing_ok=True)
    dict_read = DictReader.read(source_file, includes=includes)
    # Execute
    DictWriter.write(dict_read, target_file)
    dict_reread = DictReader.read(target_file)
    # Assert
    assert target_file.exists()
    assert dict_read == dict_reread
    assert str(dict_read) == str(dict_reread)
    # Clean up
    target_file.unlink()


class TestCreateTargetFileName:
    @pytest.mark.parametrize(
        "file_ending",
        [
            "",
            "abc",
            "cpp",
            "foam",
            "json",
            "xml",
        ],
    )
    def test_file_ending(self, file_ending: str) -> None:
        # Prepare
        file_name = f"someDictFile.{file_ending}" if file_ending else "someDictFile"
        source_file = Path(file_name)
        assert_file = Path(file_name)
        # Execute
        target_file = create_target_file_name(source_file)
        # Assert
        assert target_file == assert_file

    @pytest.mark.parametrize(
        ("output_format", "file_ending"),
        [
            ("cpp", ""),
            ("foam", "foam"),
            ("json", "json"),
            ("xml", "xml"),
            ("", ""),
            ("notdefinedformat", ""),
        ],
    )
    def test_output_format_file_ending(self, output_format: str, file_ending: str) -> None:
        # Prepare
        source_file = Path("someDictFile")
        assert_file_name = f"someDictFile.{file_ending}" if file_ending else "someDictFile"
        assert_file = Path(assert_file_name)
        # Execute
        target_file = create_target_file_name(source_file, output=output_format)
        # Assert
        assert target_file == assert_file

    def test_target_file_does_not_contain_double_prefix(self) -> None:
        # Prepare
        source_file = Path("test_dictWriter_dict")
        target_file_expected = Path("prefix.test_dictWriter_dict")
        # Execute
        target_file_first_call = create_target_file_name(source_file, "prefix")
        target_file_second_call = create_target_file_name(target_file_first_call, "prefix")
        # Assert
        assert target_file_first_call == target_file_expected
        assert target_file_second_call == target_file_expected
